/*
 *  Copyright (c) 2000-2003 Barak Weichselbaum <barak@komodia.com>
 *  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * Contact info:
 * -------------
 *
 * Site:					http://www.komodia.com
 * Main contact:			barak@komodia.com
 * For custom projects, 
 * consulting, or other
 * paid services:			sales@komodia.com
 */

#include "stdafx.h"
#include "ICMPSocket.h"

#include "ErrorHandlerMacros.h"

#include <memory>

#ifdef _MEMORY_DEBUG 
	#define new	   DEBUG_NEW  
	#define malloc DEBUG_MALLOC  
    static char THIS_FILE[] = __FILE__;  
#endif

KOMODIA_NAMESPACE_START

#define CICMPSocket_Class "CICMPSocket"

CICMPSocket::CICMPSocket(BOOL bRawSocket) : CSpoofSocket(bRawSocket),
											m_pICMPHeader(NULL),
											m_pIPHeader(NULL),
											m_pICMPIPHeader(NULL),
											m_pSendBuffer(NULL),
											m_pICMPICMPHeader(NULL),
											m_pICMPReceiveData(NULL)
{
	try
	{
		//Set our name
		SetName(CICMPSocket_Class);
	}
	ERROR_HANDLER("CICMPSocket")
}

CICMPSocket::~CICMPSocket()
{
	try
	{
		//Delete structures
		delete m_pIPHeader;
		delete m_pICMPHeader;
		delete m_pICMPIPHeader;
		delete m_pICMPICMPHeader;

		//Free form data
		delete [] m_pICMPReceiveData;
	}
	ERROR_HANDLER("~CICMPSocket")
}

BOOL CICMPSocket::Create()
{
	try
	{
		//Set our protocol
		SetProtocol(IPPROTO_ICMP);

		//Create the socket
		return CSpoofSocket::Create(IPPROTO_ICMP);
	}
	ERROR_HANDLER_RETURN("Create",FALSE)
}

BOOL CICMPSocket::SendUnreachable(const std::string& rDestinationAddress,
								  unsigned char ucType)
{
	try
	{
		return SendUnreachable(StringToLong(rDestinationAddress),
							   ucType);
	}
	ERROR_HANDLER_RETURN("SendUnreachable",FALSE)
}

BOOL CICMPSocket::SendUnreachable(IP aDestinationAddress,
								  unsigned char ucType)
{
	try
	{
		return Send(aDestinationAddress,
					ICMP_Unreachable,
					ucType);
	}
	ERROR_HANDLER_RETURN("SendUnreachable",FALSE)
}

LPICMPHeader CICMPSocket::ConstructICMP() const
{
	try
	{
		//Constructs a basic ICMP header
		LPICMPHeader lpHead;
		lpHead=new ICMPHeader;

		//Set all as zeros
		memset(lpHead,
			   0,
			   ICMPHeaderLength);

		//Set the timestamp
		lpHead->ulICMP_Originate_Timestamp=GetTickCount();

		//Return it
		return lpHead;
	}
	ERROR_HANDLER_RETURN("ConstructICMP",FALSE)
}

BOOL CICMPSocket::SendTime(const std::string& rDestinationAddress, 
						   unsigned char ucType)
{
	try
	{
		return SendTime(StringToLong(rDestinationAddress),
						ucType);
	}
	ERROR_HANDLER_RETURN("SendTime",FALSE)
}

BOOL CICMPSocket::SendTime(IP aDestinationAddress, 
						   unsigned char ucType)
{
	try
	{
		return Send(aDestinationAddress,
					ICMP_Time,
					ucType);
	}
	ERROR_HANDLER_RETURN("SendTime",FALSE)
}

BOOL CICMPSocket::Send(IP aDestinationAddress, 
					   unsigned char ucICMP, 
					   unsigned char ucType)
{
	try
	{
		//Generic ICMP send
		LPICMPHeader lpHead;
		lpHead=ConstructICMP();

		if (!lpHead)
		{
			//Report it
			ReportError("Send","Failed to construct ICMP header!");

			//Exit
			return FALSE;
		}

		//Protect it
		std::auto_ptr<ICMPHeader> pProtection(lpHead);

		//Set the values
		lpHead->ucICMPType=ucICMP;
		lpHead->ucICMPCode=ucType;

		//And the checksum
		lpHead->usICMPChecksum=CalculateChecksum((unsigned short*)lpHead,ICMPHeaderLength);

		//Send it
		BOOL bSend;
		bSend=SendICMP(lpHead,
					   aDestinationAddress);

		//Exit
		return bSend;
	}
	ERROR_HANDLER_RETURN("Send",FALSE)
}

BOOL CICMPSocket::SendParameter(const std::string& rDestinationAddress, 
								unsigned char ucError)
{
	try
	{
		return SendParameter(StringToLong(rDestinationAddress),
							 ucError);
	}
	ERROR_HANDLER_RETURN("SendParameter",FALSE)
}

BOOL CICMPSocket::SendParameter(IP aDestinationAddress, 
								unsigned char ucError)
{
	try
	{
		//Construct the header
		LPICMPHeader lpHead;
		lpHead=ConstructICMP();

		if (!lpHead)
		{
			//Report it
			ReportError("SendParameter","Failed to construct ICMP header!");

			//Exit
			return FALSE;
		}

		//Protect it
		std::auto_ptr<ICMPHeader> pProtection(lpHead);

		//Set the values
		lpHead->ucICMPType=ICMP_Parameter;
		lpHead->ucICMPCode=ICMP_Parameter_ERROR;
		lpHead->sICMP.sUC.uc1=ucError;

		//And the checksum
		lpHead->usICMPChecksum=CalculateChecksum((unsigned short*)lpHead,ICMPHeaderLength);

		//Send it
		BOOL bSend;
		bSend=SendICMP(lpHead,
					   aDestinationAddress);

		//Done
		return bSend;
	}
	ERROR_HANDLER_RETURN("SendParameter",FALSE)
}

BOOL CICMPSocket::SendQuench(const std::string& rDestinationAddress)
{
	try
	{
		return SendQuench(StringToLong(rDestinationAddress));
	}
	ERROR_HANDLER_RETURN("SendQuench",FALSE)
}

BOOL CICMPSocket::SendQuench(IP aDestinationAddress)
{
	try
	{
		return Send(aDestinationAddress,
					ICMP_Quench,
					0);
	}
	ERROR_HANDLER_RETURN("SendQuench",FALSE)
}

BOOL CICMPSocket::SendRedirect(const std::string& rDestinationAddress, 
							   unsigned char ucType, 
							   const std::string& rGatewayAddress)
{
	try
	{
		return SendRedirect(StringToLong(rDestinationAddress),
							ucType,
							StringToLong(rGatewayAddress));
	}
	ERROR_HANDLER_RETURN("SendRedirect",FALSE)
}

BOOL CICMPSocket::SendRedirect(IP aDestinationAddress, 
							   unsigned char ucType, 
							   IP aGatewayAddress)
{
	try
	{
		LPICMPHeader lpHead;
		lpHead=ConstructICMP();

		if (!lpHead)
		{
			//Report it
			ReportError("SendRedirect","Failed to construct ICMP header!");

			//Exit
			return FALSE;
		}

		//Protect it
		std::auto_ptr<ICMPHeader> pProtection(lpHead);

		//Set the values
		lpHead->ucICMPType=ICMP_Redirect;
		lpHead->ucICMPCode=ucType;
		lpHead->sICMP.sUL=aGatewayAddress;

		//And the checksum
		lpHead->usICMPChecksum=CalculateChecksum((unsigned short*)lpHead,ICMPHeaderLength);

		//Send it
		BOOL bSend;
		bSend=SendICMP(lpHead,
					   aDestinationAddress);

		//Exit
		return bSend;
	}
	ERROR_HANDLER_RETURN("SendRedirect",FALSE)
}

BOOL CICMPSocket::SendEcho(const std::string& rDestinationAddress, 
						   BOOL bReply, 
						   unsigned short usIdentifier, 
						   unsigned short usSequence, 
						   unsigned long ulData)
{
	try
	{
		return SendEcho(StringToLong(rDestinationAddress),
						bReply,
						usIdentifier,
						usSequence,
						ulData);
	}
	ERROR_HANDLER_RETURN("SendEcho",FALSE)
}

BOOL CICMPSocket::SendEcho(IP aDestinationAddress, 
						   BOOL bReply, 
						   unsigned short usIdentifier, 
						   unsigned short usSequence, 
						   unsigned long ulData)
{
	try
	{
		//Create the header
		LPICMPHeader lpHead;
		lpHead=ConstructICMP();
		
		if (!lpHead)
		{
			//Report it
			ReportError("SendEcho","Failed to construct ICMP header!");

			//Exit
			return FALSE;
		}

		//Protect it
		std::auto_ptr<ICMPHeader> pProtection(lpHead);

		//Check if echo or reply
		if (bReply)
			lpHead->ucICMPType=ICMP_Echo_Reply;
		else
			lpHead->ucICMPType=ICMP_Echo;

		lpHead->ucICMPCode=0;
		lpHead->sICMP.sUS.us1=htons(usIdentifier);
		lpHead->sICMP.sUS.us2=htons(usSequence);
		lpHead->ulICMP_Originate_Timestamp=htonl(ulData);

		//And the checksum
		lpHead->usICMPChecksum=CalculateChecksum((unsigned short*)lpHead,ICMPHeaderLength);

		//Send it
		BOOL bSend;
		bSend=CSpoofSocket::Send(aDestinationAddress,
								 (char*)lpHead,
								 ICMPHeaderLength);
		
		//Exit
		return bSend;
	}
	ERROR_HANDLER_RETURN("SendEcho",FALSE)
}

BOOL CICMPSocket::SendTimestamp(const std::string& rDestinationAddress, 
								BOOL bReply, 
								unsigned short usIdentifier, 
								unsigned short usSequence, 
								unsigned long ulOriginateTimestamp, 
								unsigned long ulReceiveTimestamp, 
								unsigned long ulTransmitTimestamp)
{
	try
	{
		return SendTimestamp(StringToLong(rDestinationAddress),
							 bReply,
							 usIdentifier,
							 usSequence,
							 ulOriginateTimestamp, 
							 ulReceiveTimestamp, 
							 ulTransmitTimestamp);
	}
	ERROR_HANDLER_RETURN("SendTimestamp",FALSE)
}

BOOL CICMPSocket::SendTimestamp(IP aDestinationAddress, 
								BOOL bReply, 
								unsigned short usIdentifier, 
								unsigned short usSequence, 
								unsigned long ulOriginateTimestamp, 
								unsigned long ulReceiveTimestamp, 
								unsigned long ulTransmitTimestamp)
{
	try
	{
		//Create the header
		LPICMPHeader lpHead;
		lpHead=ConstructICMP();

		if (!lpHead)
		{
			//Report it
			ReportError("SendTimestamp","Failed to construct ICMP header!");

			//Exit
			return FALSE;
		}

		//Protect it
		std::auto_ptr<ICMPHeader> pProtection(lpHead);

		//Check if echo or reply
		if (bReply)
			lpHead->ucICMPType=ICMP_Timestamp_Reply;
		else
			lpHead->ucICMPType=ICMP_Timestamp;

		lpHead->ucICMPCode=0;
		lpHead->sICMP.sUS.us1=htons(usIdentifier);
		lpHead->sICMP.sUS.us2=htons(usSequence);
		lpHead->ulICMP_Originate_Timestamp=htonl(ulOriginateTimestamp);
		lpHead->ulICMP_Receive_Timestamp=htonl(ulReceiveTimestamp);
		lpHead->ulICMP_Transmit_Timestamp=htonl(ulTransmitTimestamp);

		//And the checksum
		lpHead->usICMPChecksum=CalculateChecksum((unsigned short*)lpHead,ICMPHeaderLength);

		//Send it
		BOOL bSend;
		bSend=SendICMP(lpHead,
					   aDestinationAddress);

		//Exit
		return bSend;
	}
	ERROR_HANDLER_RETURN("SendTimestamp",FALSE)
}

BOOL CICMPSocket::SendInformation(const std::string& rDestinationAddress,
								  BOOL bReply,
								  unsigned short usIdentifier,
								  unsigned short usSequence)
{
	try
	{
		return SendInformation(StringToLong(rDestinationAddress),
							   bReply,
							   usIdentifier,
							   usSequence);
	}
	ERROR_HANDLER_RETURN("SendInformation",FALSE)
}

BOOL CICMPSocket::SendInformation(IP aDestinationAddress, 
								  BOOL bReply, 
								  unsigned short usIdentifier, 
								  unsigned short usSequence)
{
	try
	{
		//Create the header
		LPICMPHeader lpHead;
		lpHead=ConstructICMP();

		if (!lpHead)
		{
			//Report it
			ReportError("SendInformation","Failed to construct ICMP header!");

			//Exit
			return FALSE;
		}

		//Check if echo or reply
		if (bReply)
			lpHead->ucICMPType=ICMP_Information_Reply;
		else
			lpHead->ucICMPType=ICMP_Information;

		lpHead->ucICMPCode=0;
		lpHead->sICMP.sUS.us1=htons(usIdentifier);
		lpHead->sICMP.sUS.us2=htons(usSequence);
		
		//And the checksum
		//Using only first 8 bytes
		lpHead->usICMPChecksum=CalculateChecksum((unsigned short*)lpHead,
												 ICMP_Information_SIZE);

		//Send it
		BOOL bSend;
		bSend=SendICMP(lpHead,
					   aDestinationAddress);

		//Exit
		return bSend;
	}
	ERROR_HANDLER_RETURN("SendInformation",FALSE)
}


BOOL CICMPSocket::ProccessICMP(const char* pBuffer)
{
	try
	{
		//Here we proccess the input we received
		//Do we have previous data?
		if (m_pICMPReceiveData)
		{
			//Delete it
			delete [] m_pICMPReceiveData;
			m_pICMPReceiveData=NULL;
		}

		//No user data
		m_usDataSize=0;

		//Initialize members
		if (!m_pIPHeader)
			m_pIPHeader=new IpHeader;

		if (!m_pICMPHeader)
			m_pICMPHeader=new ICMPHeader;

		//Create an IP header
		LPIpHeader lpHead;
		lpHead=m_pIPHeader;

		//Copy to buffer
		memcpy(lpHead,
			   pBuffer,
			   IpHeaderLength);

		//Let's check for options
		unsigned char ucHeaderSize;
		ucHeaderSize=(lpHead->ucHeaderLength_Version & 15) << 2;

		//Now check for total packet size
		unsigned short ucPacketSize;
		ucPacketSize=htons(lpHead->usTotalLength);

		//Copy data to icmp
		memset(m_pICMPHeader,
			   0,
			   ICMPHeaderLength);

		//How much to copy ?
		unsigned short ucCopy;
		ucCopy=ucPacketSize-ucHeaderSize;
		
		//Save the datasize
		m_usDataSize=ucCopy;

		if (ucCopy>ICMPHeaderLength)
			ucCopy=ICMPHeaderLength;

		//Copy the ICMP header
		memcpy(m_pICMPHeader,
			   pBuffer+ucHeaderSize,
			   ucCopy);

		//Now save the original IP
		if (!m_pICMPIPHeader)
			m_pICMPIPHeader=new IpHeader;

		//Copy the data
		memcpy(m_pICMPIPHeader,
			   pBuffer+ucHeaderSize+ICMP_DATA_SIZE,
			   IpHeaderLength);

		//Get this header size
		unsigned char ucNewHeaderSize;
		ucNewHeaderSize=(m_pICMPIPHeader->ucHeaderLength_Version & 15) << 2;

		//Do we have ICMP buffer
		if (!(m_pICMPHeader->ucICMPType!=ICMP_Echo &&
			  m_pICMPHeader->ucICMPType!=ICMP_Echo_Reply &&
			  m_pICMPHeader->ucICMPType!=ICMP_Timestamp &&
			  m_pICMPHeader->ucICMPType!=ICMP_Timestamp_Reply &&
			  m_pICMPHeader->ucICMPType!=ICMP_Information &&
			  m_pICMPHeader->ucICMPType!=ICMP_Information_Reply &&
			  m_pICMPHeader->ucICMPType!=ICMP_Time))
		{
			
			//Copy rest of data
			if (!m_pICMPICMPHeader)
				m_pICMPICMPHeader=new ICMPHeader;

			memcpy(m_pICMPICMPHeader,
				   pBuffer+ucHeaderSize+ICMP_DATA_SIZE+ucNewHeaderSize,
				   ICMP_DATA_SIZE);

			//Reverse it
			ReverseHeader(m_pICMPICMPHeader);

			//Do we have data to copy
			unsigned short usRemainingSize;
			usRemainingSize=ucHeaderSize+
							ICMP_DATA_SIZE+
							ucNewHeaderSize+
							ICMP_DATA_SIZE;

			//Anything left?
			if (usRemainingSize<ucPacketSize)
			{
				//Set the size
				m_usDataSize=ucPacketSize-usRemainingSize;

				//Allocate some data
				m_pICMPReceiveData=new char[m_usDataSize];

				//Copy it
				memcpy(m_pICMPReceiveData,
					   pBuffer+usRemainingSize,
					   m_usDataSize);
			}
		}
		else
		{
			//Reset the ICMP
			if (m_pICMPICMPHeader)
				memset(m_pICMPICMPHeader,
					   0,
					   ICMPHeaderLength);

			//Do we have data to copy
			unsigned short usRemainingSize;
			usRemainingSize=ucHeaderSize+
							ICMP_DATA_SIZE+
							ucNewHeaderSize;

			//Anything left?
			if (usRemainingSize<ucPacketSize)
			{
				//Set the size
				m_usDataSize=ucPacketSize-usRemainingSize;

				//Allocate some data
				m_pICMPReceiveData=new char[m_usDataSize];

				//Copy it
				memcpy(m_pICMPReceiveData,
					   pBuffer+usRemainingSize,
					   m_usDataSize);
			}
		}

		//Now I need to reverse the header
		ReverseHeader(m_pICMPHeader);

		//Done
		return TRUE;
	}
	ERROR_HANDLER_RETURN("ProccessICMP",FALSE)
}

const LPICMPHeader CICMPSocket::GetLastICMPHeader() const
{
	//Return the last header proccessed
	return m_pICMPHeader;
}

const LPIpHeader CICMPSocket::GetLastIPHeader() const
{
	return m_pIPHeader;
}

unsigned long CICMPSocket::GetLastDataSize()
{
	return m_usDataSize;
}

void CICMPSocket::ReverseHeader(LPICMPHeader pHeader)
{
	try
	{
		//Reverse timestamps
		if (pHeader->ucICMPType==ICMP_Timestamp || pHeader->ucICMPType==ICMP_Timestamp_Reply)
		{
			pHeader->ulICMP_Originate_Timestamp=htonl(pHeader->ulICMP_Originate_Timestamp);
			pHeader->ulICMP_Receive_Timestamp=htonl(pHeader->ulICMP_Receive_Timestamp);
			pHeader->ulICMP_Transmit_Timestamp=htonl(pHeader->ulICMP_Transmit_Timestamp);
		}


		//Reverse ID and Sequence
		if (pHeader->ucICMPType==ICMP_Echo || 
			pHeader->ucICMPType==ICMP_Echo_Reply ||
			pHeader->ucICMPType==ICMP_Time)
		{
			pHeader->sICMP.sUS.us1=htons(pHeader->sICMP.sUS.us1);
			pHeader->sICMP.sUS.us2=htons(pHeader->sICMP.sUS.us2);
		}
	}
	ERROR_HANDLER("ReverseHeader")
}

const LPIpHeader CICMPSocket::GetLastICMPIPHeader() const
{
	//Get the IP header received via the icmp
	return m_pICMPIPHeader;
}

const LPICMPHeader CICMPSocket::GetLastICMPICMPHeader()const
{
	return m_pICMPICMPHeader;
}

BOOL CICMPSocket::SendICMP(LPICMPHeader lpHead, 
						   const std::string& rDestinationAddress)
{
	try
	{
		return SendICMP(lpHead,
						StringToLong(rDestinationAddress));
	}
	ERROR_HANDLER_RETURN("SendICMP",FALSE)
}

BOOL CICMPSocket::SendICMP(LPICMPHeader lpHead, 
						   IP aDestinationAddress)
{
	try
	{
		//Our buffer to send
		char* pSendBuffer;
		
		//Size of the buffer to send
		int iSendSize;
		iSendSize=ICMPHeaderLength;

		//Do we have an external buffer ?
		if (!m_pSendBuffer)			
			pSendBuffer=(char*)lpHead;
		else
		{
			iSendSize+=m_usBufferSendSize;
			pSendBuffer=new char[m_usBufferSendSize];

			//Copy the data
			memcpy(pSendBuffer,
				   (char*)lpHead,
				   ICMPHeaderLength);
			memcpy(pSendBuffer+ICMPHeaderLength,
				   m_pSendBuffer,
				   m_usBufferSendSize);
		}

		//Send to modify before send
		FinalICMPHeader((LPICMPHeader)pSendBuffer);

		//Send it
		int iResult;
		iResult=CSpoofSocket::Send(aDestinationAddress,
								   pSendBuffer,
								   iSendSize);

		if (m_pSendBuffer)
			delete [] pSendBuffer;

		//Done
		return iResult>0;
	}
	ERROR_HANDLER_RETURN("SendICMP",FALSE)
}

void CICMPSocket::SetSendBuffer(const char *pBuffer, 
								unsigned short usSendSize)
{
	//Set the buffer
	m_pSendBuffer=pBuffer;
	m_usBufferSendSize=usSendSize;
}

void CICMPSocket::FinalICMPHeader(LPICMPHeader lpHead)const
{
}

const char* CICMPSocket::GetSendBuffer()const
{
	return m_pSendBuffer;
}

int CICMPSocket::GetBufferSize()const
{
	return m_usBufferSendSize;
}

const char* CICMPSocket::GetLastData()const
{
	return m_pICMPReceiveData;
}

KOMODIA_NAMESPACE_END
