/*
 *  Copyright (c) 2000-2003 Barak Weichselbaum <barak@komodia.com>
 *  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * Contact info:
 * -------------
 *
 * Site:					http://www.komodia.com
 * Main contact:			barak@komodia.com
 * For custom projects, 
 * consulting, or other
 * paid services:			sales@komodia.com
 */

#include "stdafx.h"
#include "DNSAnswer.h"

#include "ErrorHandlerMacros.h"
#include "Array_ptr.h"

#include "SpoofBase.h"

#include "DNSParser.h"

#ifdef _MEMORY_DEBUG 
	#define new	   DEBUG_NEW  
	#define malloc DEBUG_MALLOC  
    static char THIS_FILE[] = __FILE__;  
#endif

namespace KomodiaDNS
{

#define CDNSAnswer_Class "CDNSAnswer"

CDNSAnswer::CDNSAnswer() : CErrorHandler(),
						   m_aAnswerType(dhNone),
						   m_usMXPriority(0),
						   m_pSOA(NULL)
{
	try
	{
		//Set our name
		SetName(CDNSAnswer_Class);
	}
	ERROR_HANDLER("CDNSAnswer")
}

CDNSAnswer::CDNSAnswer(const CDNSAnswer& rAnswer) : CErrorHandler(rAnswer),
												    m_aAnswerType(rAnswer.m_aAnswerType),
												    m_usMXPriority(rAnswer.m_usMXPriority),
												    m_pSOA(NULL),
													m_aAnswer(rAnswer.m_aAnswer)
{
	try
	{
		//Set our name
		SetName(CDNSAnswer_Class);

		//Do we have SOA record?
		if (rAnswer.m_pSOA)
			//Copy it
			m_pSOA=new DNSSOARecord(*rAnswer.m_pSOA);
	}
	ERROR_HANDLER("CDNSAnswer")
}

CDNSAnswer::~CDNSAnswer()
{
	try
	{
		//Delete the data
		DeleteBuffer();
	}
	ERROR_HANDLER("~CDNSAnswer")
}

CDNSAnswer& CDNSAnswer::operator=(const CDNSAnswer& rAnswer)
{
	//Check if we are the same?
	if (this==&rAnswer)
		return *this;

	//Start to copy the data
	m_aAnswerType=rAnswer.m_aAnswerType;
	m_aAnswer=rAnswer.m_aAnswer;
	m_usMXPriority=rAnswer.m_usMXPriority;
	m_aAnswer=rAnswer.m_aAnswer;

	//Delete the SOA
	delete m_pSOA;

	//Do we need to copy it
	if (rAnswer.m_pSOA)
		m_pSOA=new DNSSOARecord(*rAnswer.m_pSOA);
	else
		m_pSOA=NULL;

	//Done
	return *this;
}

const DnsRRHeader& CDNSAnswer::GetAnswerStruct()const
{
	return m_aAnswer;
}

int CDNSAnswer::ParseBuffer(const char* pBuffer,
							const char* pOriginalBuffer)
{
	try
	{
		//Parse the buffer returns number of byte proccesed
		//Copy buffer
		const char* pBackupBuffer=pBuffer;

		//Parse the name
		unsigned short usLength;
		m_aAnswer.sName=CDNSParser::ParseName(pBuffer,
											  pOriginalBuffer,
											  usLength);

		if (!usLength)
		{
			//Report it
			ReportError("ParseBuffer","Received zero length!");

			//Exit
			return 0;
		}
		
		//Increase the buffer
		pBackupBuffer+=usLength;

		//More fields
		memcpy(&m_aAnswer.usType,
			   pBackupBuffer,
			   sizeof(m_aAnswer.usType));
		m_aAnswer.usType=htons(m_aAnswer.usType);
		pBackupBuffer+=sizeof(m_aAnswer.usType);

		//More fields
		memcpy(&m_aAnswer.usClass,
			   pBackupBuffer,
			   sizeof(m_aAnswer.usClass));
		m_aAnswer.usClass=htons(m_aAnswer.usClass);
		pBackupBuffer+=sizeof(m_aAnswer.usClass);

		//More fields
		memcpy(&m_aAnswer.ulTTL,
			   pBackupBuffer,
			   sizeof(m_aAnswer.ulTTL));
		m_aAnswer.ulTTL=htonl(m_aAnswer.ulTTL);
		pBackupBuffer+=sizeof(m_aAnswer.ulTTL);

		//More fields
		memcpy(&m_aAnswer.usRdlLength,
			   pBackupBuffer,
			   sizeof(m_aAnswer.usRdlLength));
		m_aAnswer.usRdlLength=htons(m_aAnswer.usRdlLength);
		pBackupBuffer+=sizeof(m_aAnswer.usRdlLength);

		//Error indicator
		BOOL bError;
		bError=FALSE;

		//Our parsed name
		std::string sParsedName;

		//Parse the last name
		if (m_aAnswer.usType==CDNSQuery::CNAME || 
			m_aAnswer.usType==CDNSQuery::NS ||
			m_aAnswer.usType==CDNSQuery::PTR)
		{
			//Canonical name
			sParsedName=CDNSParser::ParseName(pBackupBuffer,
											  pOriginalBuffer,
											  usLength);

			//Did we have an error
			if (!usLength)
			{
				//Report it
				ReportError("ParseBuffer","Failed to parsed canonical name!");

				//Set the error
				bError=TRUE;
			}

			//Set the length
			usLength=m_aAnswer.usRdlLength;
		}
		else if (m_aAnswer.usType==CDNSQuery::A)
		{
			//Set the length
			usLength=m_aAnswer.usRdlLength;

			//Is it more then long
			if (m_aAnswer.usRdlLength>4)
			{
				//Report it
				ReportError("ParseBuffer","Unexpected A length!",m_aAnswer.usRdlLength);

				//Set the error
				bError=TRUE;
			}
			else
			{
				//Take the long
				long lIP;
				lIP=*((long*)pBackupBuffer);

				//Convert it to string
				char aTmp[11];
				sprintf(aTmp,"%lu",lIP);

				//Put in the string
				sParsedName=aTmp;
			}
		}
		else if (m_aAnswer.usType==CDNSQuery::MX)
		{
			//Read the first 2 bytes
			memcpy(&m_usMXPriority,
				   pBackupBuffer,
				   sizeof(m_usMXPriority));

			//Invert it
			m_usMXPriority=htons(m_usMXPriority);

			//Read on
			//Canonical name
			sParsedName=CDNSParser::ParseName(pBackupBuffer+sizeof(m_usMXPriority),
											  pOriginalBuffer,
											  usLength);

			//Did we have an error
			if (!usLength)
			{
				//Report it
				ReportError("ParseBuffer","Failed to parsed MX name!");

				//Set the error
				bError=TRUE;
			}

			//Set the length
			usLength=m_aAnswer.usRdlLength;
		}
		else if (m_aAnswer.usType==CDNSQuery::SOA)
		{
			//Parse it
			m_pSOA=ParseSOA(m_aAnswer.usRdlLength,
							pBackupBuffer,
							pOriginalBuffer);

			//Set the length
			usLength=m_aAnswer.usRdlLength;
		}
		else
		{
			//Report it
			ReportError("ParseBuffer","Not supported address type!");

			//Done
			return 0;
		}

		//Did we have an error
		if (bError)
			ReportError("ParseBuffer","Received zero length, trying to recover!");

		//Incease our position
		pBackupBuffer+=usLength;

		//Copy the data
		m_aAnswer.sData=sParsedName;

		//Number of bytes used
		return pBackupBuffer-pBuffer;
	}
	ERROR_HANDLER_RETURN("ParseBuffer",0)
}

void CDNSAnswer::DeleteBuffer()
{
	try
	{
		//Delete the SOA
		delete m_pSOA;
		m_pSOA=NULL;
	}
	ERROR_HANDLER("DeleteBuffer")
}

void CDNSAnswer::SetAnswerType(DnsHeaderType aType)
{
	m_aAnswerType=aType;
}

CDNSAnswer::DnsHeaderType CDNSAnswer::GetAnswerType()const
{
	return m_aAnswerType;
}

CDNSQuery::DNSQueryTypes CDNSAnswer::GetResourceType()const
{
	return (CDNSQuery::DNSQueryTypes)m_aAnswer.usType;
}

DNSSOARecord* CDNSAnswer::ParseSOA(unsigned short usSOASize,
								   const char* pBuffer,
								   const char* pOriginalBuffer)
{
	try
	{
		//Backup buffer
		const char* pBackupBuffer=pBuffer;

		//The record
		DNSSOARecord* pRecord;
		pRecord=new DNSSOARecord;

		//Protect it
		std::auto_ptr<DNSSOARecord> pProtection(pRecord);

		//Start parsing the names
		unsigned short usLength;
		pRecord->sPrimaryNameServer=CDNSParser::ParseName(pBuffer,
														  pOriginalBuffer,
														  usLength);
		//Check the size
		if (!usLength)
		{
			//Report it
			ReportError("ParseSOA","Failed to parse primary name server!");

			//Exit
			return NULL;
		}

		//Increase our position
		pBuffer+=usLength;

		//Try to parse
		pRecord->sAuthorativeMailbox=CDNSParser::ParseName(pBuffer,
														   pOriginalBuffer,
														   usLength);

		//Check the size
		if (!usLength)
		{
			//Report it
			ReportError("ParseSOA","Failed to parse authorative mailbox!");

			//Exit
			return NULL;
		}

		//Set backup buffer
		pBackupBuffer+=usSOASize-DNSSOARecordCopySize;

		//Copy rest of the data
		memcpy(&pRecord->ulVersionNumber,
			   pBackupBuffer,
			   DNSSOARecordCopySize);

		//Reverse the header
		pRecord->ulVersionNumber=htonl(pRecord->ulVersionNumber);
		pRecord->ulRefreshInterval=htonl(pRecord->ulRefreshInterval);
		pRecord->ulRetryInterval=htonl(pRecord->ulRetryInterval);
		pRecord->ulExpirationLimit=htonl(pRecord->ulExpirationLimit);
		pRecord->ulMinimumTTL=htonl(pRecord->ulMinimumTTL);

		//Release the protection
		pProtection.release();

		//Done
		return pRecord;
	}
	ERROR_HANDLER_RETURN("ParseSOA",NULL)
}

unsigned short CDNSAnswer::GetMXPriority()const
{
	return m_usMXPriority;
}

const DNSSOARecord* CDNSAnswer::GetSOARecord()const
{
	return m_pSOA;
}

const std::string& CDNSAnswer::GetData()const
{
	return m_aAnswer.sData;
}

const std::string& CDNSAnswer::GetName()const
{
	return m_aAnswer.sName;
}

std::string CDNSAnswer::GetAnswerParsed(BOOL bWindowsLines,
										BOOL bBrief)const
{
	try
	{
		//Line terminator
		//End of line
		char aEndOfLine[3];

		//Which is it
		if (bWindowsLines)
		{
			aEndOfLine[0]=13;
			aEndOfLine[1]=10;
			aEndOfLine[2]=0;
		}
		else
		{
			aEndOfLine[0]='\n';
			aEndOfLine[1]=0;
		}

		//Our answer
		std::string sAnswer;

		//The answer
		//What are we
		if (m_aAnswer.usType==CDNSQuery::SOA)
		{
			//Our name
			sAnswer+=m_aAnswer.sName;
			sAnswer+=aEndOfLine;

			//Dump the soa data
			sAnswer+=GetSOARecord(std::string("    "),
								  aEndOfLine);
		}
		if (m_aAnswer.usType==CDNSQuery::A)
		{
			//Regular address
			//Is it an address
			if (m_aAnswerType==dhAnswer)
			{
				//Our name
				sAnswer+="Name: ";
				sAnswer+=m_aAnswer.sName;
				sAnswer+=aEndOfLine;

				//Address
				sAnswer+="Address: ";
			}
			else if (m_aAnswerType==dhAdditional)
			{
				//Our name
				sAnswer+=m_aAnswer.sName;
				sAnswer+=' ';

				//Address prefix
				sAnswer+="internet address = ";
			}

			//The address
			sAnswer+=CSpoofBase::LongToStdString(atol(m_aAnswer.sData.c_str()));
		}
		else if (m_aAnswer.usType==CDNSQuery::MX)
		{
			//Our name
			sAnswer+=m_aAnswer.sName;
			sAnswer+=' ';

			//Convert the prefrence
			char aTmp[11];
			itoa(m_usMXPriority,aTmp,10);

			//MX prefrence
			sAnswer+="MX preference = ";
			sAnswer+=aTmp;
			sAnswer+=std::string(", ");

			//MX address
			sAnswer+="mail exchanger = ";
			sAnswer+=m_aAnswer.sData;
		}
		else if (m_aAnswer.usType==CDNSQuery::NS)
		{
			//Our name
			sAnswer+=m_aAnswer.sName;
			sAnswer+=' ';

			//The address
			sAnswer+="nameserver = ";
			sAnswer+=m_aAnswer.sData;
		}
		else if (m_aAnswer.usType==CDNSQuery::PTR)
		{
			//Our name
			sAnswer+=m_aAnswer.sName;
			sAnswer+=' ';

			//The address
			sAnswer+="name = ";
			sAnswer+=m_aAnswer.sData;
		}

		//Done
		return sAnswer;
	}
	ERROR_HANDLER_RETURN("GetAnswerParsed","")
}

std::string CDNSAnswer::GetSOARecord(const std::string& rShift,
									 const std::string& rTerminator)const
{
	try
	{
		//Our string
		std::string sSOA;

		//Start to add
		//Primary NS
		sSOA+=rShift;
		sSOA+="primary name server = ";
		sSOA+=m_pSOA->sPrimaryNameServer;
		sSOA+=rTerminator;

		//Primary mailbox
		sSOA+=rShift;
		sSOA+="responsible mail addr = ";
		sSOA+=m_pSOA->sAuthorativeMailbox;
		sSOA+=rTerminator;

		//Version
		sSOA+=rShift;
		sSOA+="serial  = ";
		sSOA+=ConvertUL(m_pSOA->ulVersionNumber);
		sSOA+=rTerminator;

		//Refresh
		sSOA+=rShift;
		sSOA+="refresh = ";
		sSOA+=ConvertUL(m_pSOA->ulRefreshInterval);
		sSOA+=ConvertMS(m_pSOA->ulRefreshInterval);
		sSOA+=rTerminator;

		//Retry
		sSOA+=rShift;
		sSOA+="retry   = ";
		sSOA+=ConvertUL(m_pSOA->ulRetryInterval);
		sSOA+=ConvertMS(m_pSOA->ulRetryInterval);
		sSOA+=rTerminator;

		//Expire
		sSOA+=rShift;
		sSOA+="expire  = ";
		sSOA+=ConvertUL(m_pSOA->ulExpirationLimit);
		sSOA+=ConvertMS(m_pSOA->ulExpirationLimit);
		sSOA+=rTerminator;

		//Default TTL
		sSOA+=rShift;
		sSOA+="default TTL = ";
		sSOA+=ConvertUL(m_pSOA->ulMinimumTTL);
		sSOA+=ConvertMS(m_pSOA->ulMinimumTTL);
		sSOA+=rTerminator;

		//Done
		return sSOA;
	}
	ERROR_HANDLER_RETURN("GetSOARecord","")
}

std::string CDNSAnswer::ConvertUL(unsigned long ulNumber)
{
	//Convert
	char aTmp[11];
	sprintf(aTmp,"%lu",ulNumber);

	//Done
	return aTmp;
}

std::string CDNSAnswer::ConvertMS(unsigned long ulNumber)
{
	//Our string
	char aTmp[100];

	//What are we
	if (ulNumber<60)
		sprintf(aTmp," (%lu seconds)",ulNumber);
	else if (ulNumber<3600)
		sprintf(aTmp," (%lu minutes)",ulNumber/60);
	else if (ulNumber<86400)
		sprintf(aTmp," (%lu hours)",ulNumber/3600);
	else
		sprintf(aTmp," (%lu days)",ulNumber/86400);

	//Done
	return aTmp;
}

//of namespace
}