/*
 *  Copyright (c) 2000-2003 Barak Weichselbaum <barak@komodia.com>
 *  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * Contact info:
 * -------------
 *
 * Site:					http://www.komodia.com
 * Main contact:			barak@komodia.com
 * For custom projects, 
 * consulting, or other
 * paid services:			sales@komodia.com
 */

#ifndef _DNSAnswer_H_
#define _DNSAnswer_H_

#include "ErrorHandler.h"

#include "DNSQuery.h"

namespace KomodiaDNS
{

// Header of the DNS RR
typedef struct _DnsRRHeader 
{
	std::string			sName;				// Variable
	unsigned short		usType;				// Too many fields
	unsigned short		usClass;
	unsigned long		ulTTL;
	unsigned short		usRdlLength;
	std::string			sData;				// Variable

	//Default initializing ctor
	_DnsRRHeader() : usType(0),
					 usClass(0),
					 ulTTL(0),
					 usRdlLength(0)
	{
	}
} DnsRRHeader;

typedef DnsRRHeader FAR * LPDnsRRHeader;

//SOA record
typedef struct _DNSSOARecord
{
	std::string		sPrimaryNameServer;
	std::string 	sAuthorativeMailbox;
	unsigned long	ulVersionNumber;
	unsigned long	ulRefreshInterval;
	unsigned long	ulRetryInterval;
	unsigned long	ulExpirationLimit;
	unsigned long	ulMinimumTTL;

	//Default initializing ctor
	_DNSSOARecord() : ulVersionNumber(0),
					  ulRefreshInterval(0),
					  ulRetryInterval(0),
					  ulExpirationLimit(0),
					  ulMinimumTTL(0)
	{
	}
} DNSSOARecord;

#define DNSSOARecordCopySize (sizeof(unsigned long)*5)

class CDNSAnswer : public CErrorHandler
{
public:
	//Type of headers
	typedef enum _DnsHeaderType
	{
		dhQuestion,
		dhAnswer,
		dhAuthoritive,
		dhAdditional,
		dhNone
	} DnsHeaderType;
public:
	//Create a string of the answer
	std::string GetAnswerParsed(BOOL bWindowsLines=TRUE,
								BOOL bBrief=TRUE)const;

	//Get the mail priority
	unsigned short GetMXPriority()const;

	//Get the SOA record (if any ?)
	const DNSSOARecord* GetSOARecord()const;

	//Get the answer data
	const DnsRRHeader& GetAnswerStruct()const;

	//What type of answer are we
	DnsHeaderType GetAnswerType()const;

	//Get the class
	CDNSQuery::DNSQueryTypes GetResourceType()const;

	//Get the data
	const std::string& GetData()const;

	//Get the name
	const std::string& GetName()const;

	//Set our answer type
	void SetAnswerType(DnsHeaderType aType);

	//Parse the answer
	int ParseBuffer(const char* pBuffer,
					const char* pOriginalBuffer);

	//Assignment operator
	CDNSAnswer& operator=(const CDNSAnswer& rAnswer);

	//Ctor and Dtor
	CDNSAnswer();
	CDNSAnswer(const CDNSAnswer& rAnswer);
	virtual ~CDNSAnswer();
private:
	//Parse the SOA
	DNSSOARecord* ParseSOA(unsigned short usSOASize,
						   const char* pBuffer,
						   const char* pOriginalBuffer);

	//Delete the record
	void DeleteBuffer();

	//Get the SOA record as string
	std::string GetSOARecord(const std::string& rShift,
							 const std::string& rTerminator)const;

	//Convert a number to string
	static std::string ConvertUL(unsigned long ulNumber);

	//Convert MS
	static std::string ConvertMS(unsigned long ulNumber);

	//Our answer type
	DnsHeaderType m_aAnswerType;

	//Our answer
	DnsRRHeader m_aAnswer;

	//MX priorty
	unsigned short m_usMXPriority;

	//Our SOA record
	DNSSOARecord* m_pSOA;
};

//of namespace
}

#endif