/*
 *  Copyright (c) 2000-2003 Barak Weichselbaum <barak@komodia.com>
 *  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * Contact info:
 * -------------
 *
 * Site:					http://www.komodia.com
 * Main contact:			barak@komodia.com
 * For custom projects, 
 * consulting, or other
 * paid services:			sales@komodia.com
 */

#ifndef _DNSAnswers_H_
#define _DNSAnswers_H_

#include "ErrorHandler.h"

#include "DNSAnswer.h"

#include <vector>

namespace KomodiaDNS
{

class CDNSQuery;

// Header of the DNS header
typedef struct _DnsHeaderHeader 
{
	unsigned short		usID;				// DNS header ID
	unsigned short		usOptions;			// Too many fields
	unsigned short		usQDCount;
	unsigned short		usANCount;
	unsigned short		usNSCount;
	unsigned short		usARCount;

	//Initialize ctor
	_DnsHeaderHeader() : usID(0),
						 usOptions(0),
						 usQDCount(0),
						 usANCount(0),
						 usNSCount(0),
						 usARCount(0)
	{
	}
} DnsHeaderHeader;

typedef DnsHeaderHeader FAR * LPDnsHeaderHeader;

#define DnsHeaderHeaderLength sizeof(DnsHeaderHeader)

//DNS headers variables
#define DNS_QR_REQUEST 0			//Request DNS entry
#define DNS_QR_REPLY (1 << 7)		//DNS reply

//four bit field that specifies kind of query in this message
#define DNS_OPCODE_STANDARD 0		
#define DNS_OPCODE_INVERSE (1 << 3)
#define DNS_OPCODE_STATUS (2 << 3)

//Authoritative Answer - this bit is valid in responses,
//and specifies that the responding name server is an
//authority for the domain name in question section.
#define DNS_AA (1 << 2)

//Recursion Desired - this bit may be set in a query and
//is copied into the response.  If RD is set, it directs
//the name server to pursue the query recursively.
//Recursive query support is optional.
#define DNS_RECURSION 1

//Recursion Available - this be is set or cleared in a
//response, and denotes whether recursive query support is
//available in the name server.
#define DNS_RECURSION_AVAILABLE (1 << 8)

//Response code - this 4 bit field is set as part of
//responses.  The values have the following
//interpretation
#define DNS_RCODE_NOERROR 0
#define DNS_RCODE_FORMAT (1 << 13)
#define DNS_RCODE_SERVER (2 << 13)
#define DNS_RCODE_NAME (3 << 13)
#define DNS_RCODE_REFUSED (5 << 13)

//Server failure
#define DNS_SERVER_FAILURE (3 << 8)

class CDNSAnswers : public CErrorHandler
{
public:
	//Vector of answers
	typedef std::vector<std::string> StringAnswersVector;

	//Vector of IP answers
	typedef std::vector<unsigned long> IPAnswersVector;
public:
	//Create a string of the answer
	//bWindowsLines - Use window line seperator
	//bBrief - Make the report brief
	//bExtraLine - Add an extra line seperator in the end
	std::string GetAnswerParsed(BOOL bWindowsLines=TRUE,
								BOOL bBrief=TRUE,
								BOOL bExtraLine=FALSE)const;

	//Get the query
	const CDNSQuery* GetQuery()const;

	//Is the DNS resovled ?
	BOOL IsResolved()const;

	//Is it a reverse answer
	BOOL IsReverseAnswer()const;

	//First answer data (if you expect many answers, 
	//request it with the vector)
	std::string GetAnswer()const;

	//Get all the answers
	StringAnswersVector GetAnswers()const;

	//Get the first A/IP answer (if you expect many answers, 
	//request it with the vector)
	unsigned long GetAIPAnswer()const;

	//Get all the A/IP answers
	IPAnswersVector GetAIPAnswers()const;

	//Get the request ID
	unsigned short GetDNSID() const;

	//With all this data, just get the address
	unsigned long GetAddress(unsigned short usIndex=0)const;

	//Get an answer
	const CDNSAnswer& GetAnswer(unsigned short usIndex)const;

	//Parse the answers
	int Parse(const char *pBuffer,
			  int iLength);

	//Ctor and Dtor
	CDNSAnswers();
	CDNSAnswers(const CDNSAnswers& rAnswers);
	virtual ~CDNSAnswers();
private:
	//Our vector of answers
	typedef std::vector<CDNSAnswer> AnswersVector;
private:
	//No assignment operator
	CDNSAnswers& operator=(const CDNSAnswers& rAnswers);

	//Parse the answer
	unsigned short ParseAnswers(const char *pBuffer,
								const char* pOriginalBuffer);

	//Reverse header data
	void ReverseHeader();

	//Delete all the data
	void UnInitialize();

	//Our header
	DnsHeaderHeader m_aDNSHeader;

	//The query
	CDNSQuery* m_pDNSQuery;
	
	//Our answers
	AnswersVector m_aAnswers;
};

//of namespace
}

#endif