/*
 *  Copyright (c) 2000-2003 Barak Weichselbaum <barak@komodia.com>
 *  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * Contact info:
 * -------------
 *
 * Site:					http://www.komodia.com
 * Main contact:			barak@komodia.com
 * For custom projects, 
 * consulting, or other
 * paid services:			sales@komodia.com
 */

#include "stdafx.h"
#include "DNSParser.h"

#include "ErrorHandlerMacros.h"
#include "Array_ptr.h"

#ifdef _MEMORY_DEBUG 
	#define new	   DEBUG_NEW  
	#define malloc DEBUG_MALLOC  
    static char THIS_FILE[] = __FILE__;  
#endif

namespace KomodiaDNS
{

#define CDNSParser_Class "CDNSParser"

#define MAX_REPLY 2000
#define MAX_DOMAIN_SIZE 1024

CDNSParser::CDNSParser() : CErrorHandler()
{
	try
	{
		//Set our name
		SetName(CDNSParser_Class);
	}
	ERROR_HANDLER("CDNSParser")
}

CDNSParser::~CDNSParser()
{
}

std::string CDNSParser::ParseName(const char* pBuffer,
								  const char* pOriginalBuffer,
								  unsigned short& rLength)
{
	try
	{
		//Set the length to zero
		rLength=0;

		//Damn shortcuts
		unsigned short usShort;
		usShort=*((unsigned short*)pBuffer);
		
		//Convert it
		usShort=htons(usShort);

		//Check for shortcuts
		if (usShort & 0xc000)
		{
			//Remove the traling bits
			usShort=usShort & 0x3fff;

			//Get length of the domain
			unsigned short usLength;
			usLength=GetParsedDomainLength(pOriginalBuffer+usShort,
										   pOriginalBuffer);

			//Check for errors
			if (!usLength)
			{
				//Report it
				ReportStaticError(CDNSParser_Class,"ParseName","Received zero length!");

				//Exit
				return "";
			}
			else
			{
				//Set the length to parsed
				rLength=2;

				//Returned the parsed name
				return ConvertParsedName(pOriginalBuffer+usShort,
									     pOriginalBuffer);
			}
		}
		else
		{
			//Regular name
			//Get length of the domain
			unsigned short usLength;
			usLength=GetParsedDomainLength(pBuffer,
										   pOriginalBuffer);

			//Check for errors
			if (!usLength)
			{
				//Report it
				ReportStaticError(CDNSParser_Class,"ParseName","Received zero length!");

				//Exit
				return "";
			}
			else
			{
				//Set the length
				rLength=usLength;

				//Returned the parsed data
				return ConvertParsedName(pBuffer,
									     pOriginalBuffer);
			}
		}
	}
	ERROR_HANDLER_STATIC_RETURN(CDNSParser_Class,"ParseName","")
}

unsigned short CDNSParser::GetParsedDomainLength(const char *pBuffer,
												 const char* pOriginalBuffer)
{
	try
	{
		//Get the length of the parsed domain
		const unsigned char* pBackupBuffer=(const unsigned char*)pBuffer;

		//Counter
		unsigned short usCounter;
		usCounter=0;

		while (*pBackupBuffer && 
			   usCounter<256)
		{
			if (pOriginalBuffer)
			{
				//Damn shortcuts
				unsigned short usShort;
				usShort=*((unsigned short*)pBackupBuffer);
			
				//Convert it
				usShort=htons(usShort);

				//Check for shortcuts
				if (usShort & 0xc000)
				{
					//Remove the traling bits
					usShort=usShort & 0x3fff;

					//Check the size
					if (usShort>MAX_REPLY)
					{
						//Report it
						ReportStaticError(CDNSParser_Class,"GetParsedDomainLength","Size too big!");

						//Exit
						return 0;
					}

					//Get length of the domain
					unsigned short usLength;
					usLength=GetParsedDomainLength(pOriginalBuffer+usShort,
												   pOriginalBuffer);

					//Check for errors
					if (!usLength)
					{
						//Report it
						ReportStaticError(CDNSParser_Class,"GetParsedDomainLength","Received zero length!");

						//Exit
						return 0;
					}
					else
					{
						//We are done
						usCounter+=usLength;

						//And exit
						return usCounter+1;
					}
				}
			}
			
			//Regular
			usCounter+=*pBackupBuffer+1;
			pBackupBuffer+=*pBackupBuffer+1;
		}
		
		//Check for errors
		if (usCounter>256)
		{
			//Report it
			ReportStaticError(CDNSParser_Class,"GetParsedDomainLength","Recieved length too high!");

			//Exit
			return 0;
		}
		else
			return usCounter+1;
	}
	ERROR_HANDLER_STATIC_RETURN(CDNSParser_Class,"GetParsedDomainLength",0)
}

std::string CDNSParser::ConvertParsedName(const char* pBuffer,
										  const char* pOriginalBuffer)
{
	try
	{
		//Our string
		std::string sParsedName;

		//Counter
		unsigned short usCounter;
		usCounter=0;

		while (*pBuffer && 
			   usCounter<MAX_DOMAIN_SIZE)
		{
			if (pOriginalBuffer)
			{
				//Damn shortcuts
				unsigned short usShort;
				usShort=*((const unsigned short*)pBuffer);
			
				//Convert it
				usShort=htons(usShort);

				//Check for shortcuts
				if (usShort & 0xc000)
				{
					//Remove the traling bits
					usShort=usShort & 0x3fff;

					//Check the size
					if (usShort>MAX_REPLY)
					{
						//Report it
						ReportStaticError(CDNSParser_Class,"ConvertParsedName","Size too big!");

						//Exit
						return 0;
					}

					//Get length of the domain
					unsigned short usLength;
					usLength=GetParsedDomainLength(pOriginalBuffer+usShort,
												   pOriginalBuffer);

					//Check for errors
					if (!usLength)
					{
						//Report it
						ReportStaticError(CDNSParser_Class,"ConvertParsedName","Recieved zero length!");

						//Exit
						return "";
					}
					else
						//Continue parsing
						pBuffer=pOriginalBuffer+usShort;
				}
			}
			
			//Get the section length
			unsigned short usLength;
			usLength=*pBuffer;

			//Point it to the next byte
			++pBuffer;

			//Copy it to tmp variable
			CArray_ptr<char> pTmpData(new char[usLength+1]);
			memcpy(pTmpData.Get(),
				   pBuffer,
				   usLength);

			//Terminate it
			pTmpData.Get()[usLength]=0;

			//Add it to our string
			sParsedName+=pTmpData.Get();

			//Increment the buffer
			pBuffer+=usLength;

			//Check if not the latest byte
			if (*pBuffer)
				//Add the dot
				sParsedName+='.';
		}

		//Check for errors
		if (usCounter>MAX_DOMAIN_SIZE)
		{
			//Report it
			ReportStaticError(CDNSParser_Class,"ConvertParsedName","Recieved size too large!");

			//Exit
			return "";
		}
		else
			//Done
			return sParsedName;
	}
	ERROR_HANDLER_STATIC_RETURN(CDNSParser_Class,"ConvertParsedName","")
}

std::string CDNSParser::ConvertName(const std::string& rName)
{
	try
	{
		//Length
		unsigned short usLength;

		//The string
		std::string sAnswer;
		sAnswer=ParseName(rName.c_str(),
						  rName.c_str(),
						  usLength);

		//What we have
		if (usLength)
			return sAnswer;
		else
			return "";
	}
	ERROR_HANDLER_STATIC_RETURN(CDNSParser_Class,"ConvertName","")
}

//of namespace
}