/*
 *  Copyright (c) 2000-2003 Barak Weichselbaum <barak@komodia.com>
 *  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * Contact info:
 * -------------
 *
 * Site:					http://www.komodia.com
 * Main contact:			barak@komodia.com
 * For custom projects, 
 * consulting, or other
 * paid services:			sales@komodia.com
 */

#include "stdafx.h"
#include "DNSQuery.h"

#include "ErrorHandlerMacros.h"

#include "DNSManager.h"
#include "DNSParser.h"

#ifdef _MEMORY_DEBUG 
	#define new	   DEBUG_NEW  
	#define malloc DEBUG_MALLOC  
    static char THIS_FILE[] = __FILE__;  
#endif

namespace KomodiaDNS
{

#define CDNSQuery_Class "CDNSQuery"

CDNSQuery::CDNSQuery() : CErrorHandler()
{
	try
	{
		//Set our name
		SetName(CDNSQuery_Class);
	}
	ERROR_HANDLER("CDNSQuery")
}

CDNSQuery::CDNSQuery(const CDNSQuery& rQuery) : CErrorHandler(rQuery),
												m_aQueries(rQuery.m_aQueries)
{
	try
	{
		//Set our name
		SetName(CDNSQuery_Class);
	}
	ERROR_HANDLER("CDNSQuery")
}

CDNSQuery::~CDNSQuery()
{
}

CDNSQuery& CDNSQuery::operator=(const CDNSQuery& rQuery)
{
	//Copy the vetor
	QueryVector aTmp(rQuery.m_aQueries);
	m_aQueries.swap(aTmp);

	//Done
	return *this;
}

void CDNSQuery::AddQuery(const std::string& rDomain,
						 DNSQueryTypes aType)
{
	try
	{
		//Construct the structure
		DnsQuestionHeader aData;

		//Start putting the data
		aData.sName=CDNSManager::ParseAddress(rDomain);
		aData.usType=htons(aType);
		aData.usClass=htons(DNS_CLASS_INTERNET);

		//Add to the vector
		m_aQueries.push_back(aData);
	}
	ERROR_HANDLER("AddQuery")
}

unsigned short CDNSQuery::GetQuestionCount() const
{
	return m_aQueries.size();
}

int CDNSQuery::Parse(const char* pBuffer,
					 int iNumber)
{
	try
	{
		//Our backup buffer for iterations
		const char* pBackupBuffer;
		pBackupBuffer=pBuffer;

		//Start to iterate the data
		for (int iCount=0;iCount<iNumber;++iCount)
		{
			//Try to parse the name
			std::string sParsedName;

			//The parsed length
			unsigned short usLength;

			//Parse it
			sParsedName=CDNSParser::ParseName(pBuffer,
											  pBuffer,
											  usLength);

			//Check the length
			if (!usLength)
			{
				//Report it
				ReportError("Parse","Received zero length!");

				//Exit
				return NULL;
			}

			//Add it to our header
			DnsQuestionHeader aData;

			//Set the string name
			aData.sName=sParsedName;

			//Increase backup length
			pBackupBuffer+=usLength;

			//Get the rest of the data
			memcpy(&aData.usType,
				   pBackupBuffer,
				   DnsQuestionCopyHeader);

			//Reverse the two filds
			aData.usClass=htons(aData.usClass);
			aData.usType=htons(aData.usType);

			//Add it to the question vector
			m_aQueries.push_back(aData);

			//Increase the buffer
			pBackupBuffer+=DnsQuestionCopyHeader;
		}

		//Done
		return pBackupBuffer-pBuffer;
	}
	ERROR_HANDLER_RETURN("Parse",0)
}

unsigned short CDNSQuery::GetSerializationSize()const
{
	try
	{
		//Our size counter
		unsigned short usSize;
		usSize=0;

		//Iterate the vector
		QueryVector::const_iterator aIterator;
		aIterator=m_aQueries.begin();

		//Start the loop
		while (aIterator!=m_aQueries.end())
		{
			//Calculate the struct size
			usSize+=aIterator->sName.length()+1;
			usSize+=sizeof(unsigned short)*2;

			//Next data
			++aIterator;
		}

		//Done
		return usSize;
	}
	ERROR_HANDLER_RETURN("GetSerializationSize",0)
}

unsigned short CDNSQuery::SerializeQuery(char* pBuffer)const
{
	try
	{
		//Our size counter
		unsigned short usSize;
		usSize=0;

		//Iterate the vector
		QueryVector::const_iterator aIterator;
		aIterator=m_aQueries.begin();

		//Start the loop
		while (aIterator!=m_aQueries.end())
		{
			//Calculate the struct size
			usSize+=aIterator->sName.length()+1;
			usSize+=sizeof(unsigned short)*2;

			//Copy the string
			memcpy(pBuffer,
				   aIterator->sName.c_str(),
				   aIterator->sName.length());
			pBuffer+=aIterator->sName.length();

			//Terminating null
			*pBuffer=0;
			pBuffer++;

			//Copy the data
			//Type
			memcpy(pBuffer,
				   &aIterator->usType,
				   sizeof(aIterator->usType));
			pBuffer+=sizeof(unsigned short);

			//Class
			memcpy(pBuffer,
				   &aIterator->usClass,
				   sizeof(aIterator->usClass));
			pBuffer+=sizeof(unsigned short);

			//Next data
			++aIterator;
		}

		//Done
		return usSize;
	}
	ERROR_HANDLER_RETURN("SerializeQuery",0)
}

CDNSQuery::DNSQueryTypes CDNSQuery::GetFirstQueryType()const
{
	try
	{
		//Do we have the data
		if (m_aQueries.empty())
			return ERR;
		else
			return (DNSQueryTypes)m_aQueries.front().usType;
	}
	ERROR_HANDLER_RETURN("GetFirstQueryType",ERR)
}

int CDNSQuery::GetQueryCount()const
{
	return m_aQueries.size();
}

const DnsQuestionHeader& CDNSQuery::GetQuery(int iIndex)const
{
	return m_aQueries[iIndex];
}

std::string CDNSQuery::GetQueryString(int iIndex)const
{
	try
	{
		//Get the string
		std::string sName;
		sName=m_aQueries[iIndex].sName;

		//Convert it
		return CDNSParser::ConvertName(sName);
	}
	ERROR_HANDLER_RETURN("GetQueryString","")
}

//of namespace
}