/*
 *  Copyright (c) 2000-2003 Barak Weichselbaum <barak@komodia.com>
 *  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * Contact info:
 * -------------
 *
 * Site:					http://www.komodia.com
 * Main contact:			barak@komodia.com
 * For custom projects, 
 * consulting, or other
 * paid services:			sales@komodia.com
 */

#include "stdafx.h"
#include "DNSSocket.h"

#include "ErrorHandlerMacros.h"
#include "Array_Ptr.h"
#include "OSManager.h"
#include "ThreadPool.h"
#include "GenericCriticalSection.h"

#include "DNSManager.h"
#include "DNSUDPSocket.h"
#include "DNSTCPSocket.h"

#include <algorithm>

#ifdef _MEMORY_DEBUG 
	#define new	   DEBUG_NEW  
	#define malloc DEBUG_MALLOC  
    static char THIS_FILE[] = __FILE__;  
#endif

namespace KomodiaDNS
{

#define CDNSSocket_Class "CDNSSocket"

CDNSSocket::CDNSSocket(CDNSManager* pFather,
					   const std::string strDNSServer,
					   BOOL bTCP,
					   BOOL bAsync,
					   BOOL bAsyncConnect) : CErrorHandler(),
											 m_pFather(pFather),
											 m_bUseTCP(bTCP),
											 m_bAsync(bAsync),
											 m_strDNSServer(strDNSServer),
											 m_pThreadManager(NULL),
											 m_pTCP(NULL),
											 m_pUDP(NULL),
											 m_pCSection(NULL),
											 m_bOwners(FALSE),
											 m_bAsyncConnect(bAsyncConnect),
											 m_aConnectionStatus(csNone),
											 m_pCSectionDeque(NULL)
{
	try
	{
		//Set our name
		SetName(CDNSSocket_Class);

		//The timeout
		m_iTimeout=INITIAL_TIMEOUT;

		//Create the CS
		m_pCSection=COSManager::CreateCriticalSection();
		m_pCSectionDeque=COSManager::CreateCriticalSection();
	}
	ERROR_HANDLER("CDNSSocket")
}

CDNSSocket::~CDNSSocket()
{
	try
	{
		//Deletet the data
		std::for_each(m_aPendingData.begin(),
					  m_aPendingData.end(),
					  DeleteProc);

		//Clear it
		m_aPendingData.clear();

		//Delete the sockets
		//Do we have these sockets ?
		if (m_pTCP)
			m_pTCP->DeleteSocketFromThread();

		if (m_pUDP)
			m_pUDP->DeleteSocketFromThread();

		if (m_bOwners)
			//Delete the thread manager
			delete m_pThreadManager;

		//Delete the CS
		delete m_pCSection;
		delete m_pCSectionDeque;
	}
	ERROR_HANDLER("~CDNSSocket")
}

void CDNSSocket::DeleteProc(WaitingRequests& rRequest)
{
	try
	{
		delete [] rRequest.pData;
	}
	ERROR_HANDLER_STATIC(CDNSSocket_Class,"DeleteProc")
}

BOOL CDNSSocket::Initialize()
{
	try
	{
		//Initialize our sockets
		//if TCP
		if (m_bUseTCP)
		{
			//Do we have the socket
			if (m_pTCP)
				delete m_pTCP;

			//Create the socket
			m_pTCP=new CDNSTCPSocket(this);

			//Try to create it
			if (!m_pTCP->Create())
			{
				//Write the error
				ReportError("Initialize","Failed to create socket!");

				//Exit
				return FALSE;
			}

			//Set the blocking mode
			if (!m_bAsync)
				if (!m_pTCP->ReBlock())
				{
					//Write an error
					ReportError("Initialize","Failed to switch socket to blocking mode!");

					//And exit
					return FALSE;
				}
				else
					;
			else
				//Set it to have buffer
				m_pTCP->AllowBlockedBuffer(TRUE);
		}
		else
		{
			//Do we have it
			if (m_pUDP)
				delete m_pUDP;

			//Create the socket
			m_pUDP=new CDNSUDPSocket(this);

			//Try to create it
			if (!m_pUDP->Create())
			{
				//Write the error
				ReportError("Initialize","Failed to create socket!");

				//Exit
				return FALSE;
			}

			//Set the blocking mode
			if (!m_bAsync)
				if (!m_pUDP->ReBlock())
				{
					//Write an error
					ReportError("Initialize","Failed to switch socket to blocking mode!");

					//And exit
					return FALSE;
				}
				else
					;
			else
				if (!m_pUDP->Listen())
				{
					//Write an error
					ReportError("Initialize","Failed to listen on socket to blocking mode!");

					//And exit
					return FALSE;
				}
		}

		//Done
		return TRUE;
	}
	ERROR_HANDLER_RETURN("Initialize",FALSE)
}

void CDNSSocket::SetConnectionStatus(ConnectionStatus aStatus)
{
	try
	{
		//Lock the data
		CCriticalAutoRelease aRelease(m_pCSection);

		//Save the status
		m_aConnectionStatus=aStatus;
	}
	ERROR_HANDLER("SetConnectionStatus")
}

CThreadPool* CDNSSocket::GetThreadManager()const
{
	return m_pThreadManager;
}

BOOL CDNSSocket::Send(const char *pBuffer, 
					  unsigned short bufLen,
					  LPVOID lpLocalIDData)
{
	try
	{
		//Error indicator
		int iError;
		iError=0;

		//Are we TCP
		if (m_bUseTCP)
		{
			//Reconnection status
			BOOL bReconnect;
			bReconnect=FALSE;

			//Enter the CS
			CCriticalAutoRelease aRelease(m_pCSection);

			//Are we disconnected
			if (m_aConnectionStatus==csDisconnected)
				//Try to reinitialize
				if (!Initialize())
				{
					//Write the error
					ReportError("Send","Failed to reinitialize!");

					//Done
					return FALSE;
				}
				else
					m_aConnectionStatus=csNone;

			//What kind of use ?
			if (m_aConnectionStatus==csNone)
			{
				//Set connection to pending
				SetConnectionStatus(csPending);

				//Exit the CS
				aRelease.Exit();

				//connect
				BOOL bResult;
				bResult=m_pTCP->Connect(0,
										m_strDNSServer.c_str(),
										53,
										!m_bAsyncConnect ||
										bReconnect);

				//Check for error
				if (!bResult)
				{
					//Set the connection status
					SetConnectionStatus(csNone);

					//Write the error
					ReportError("Send","Failed to connect!");

					//Get the error
					iError=m_pTCP->GetSystemLastError();
				}
				else if (!m_bAsyncConnect ||
						 !m_bAsync ||
						 bReconnect)
					//Set the connection status
					SetConnectionStatus(csConnected);
			}
			else
				aRelease.Exit();

			//Do we need to add to the pool
			if (m_aConnectionStatus!=csConnected &&
				!iError)
				//Add to queue
				AddData(pBuffer,
						bufLen,
						lpLocalIDData);
			//Can we send the data
			else if (!iError &&
					 m_aConnectionStatus==csConnected)
			{
				//Try to flush the data
				if (!Flush())
				{
					//Write the error
					ReportError("Send","Failed to flush!");

					//Get the error
					iError=1;
				}
				else if (!m_pTCP->Send(pBuffer,bufLen))
				{
					//Write the error
					ReportError("Send","Failed to send!");

					//Get the error
					iError=m_pTCP->GetSystemLastError();
				}
			}
		}
		else if (m_pUDP->Send(0,
				 m_strDNSServer.c_str(),
				 53,
				 pBuffer,
				 bufLen)==CSpoofBase::GetErrorCode())
		{
			//Write the error
			ReportError("Send","Failed to send UDP packet!");

			//Get the error
			iError=m_pUDP->GetSystemLastError();
		}

		//Did we have an error ?
		if (iError)
			//Exit
			return FALSE;
		else
			//Parse the data
			return TRUE;
	}
	ERROR_HANDLER_RETURN("Send",FALSE)
}

void CDNSSocket::SocketConnected(int iErrorCode)
{
	try
	{
		{
			//Enter the CS
			CCriticalAutoRelease aRelease(m_pCSection);

			if (iErrorCode)
				//Set the connection status
				SetConnectionStatus(csError);
			else
				SetConnectionStatus(csConnected);
		}

		//Flush the data
		Flush();
	}
	ERROR_HANDLER("SocketConnected")
}

void CDNSSocket::AddData(const char* pData,
		   				 unsigned short usSize,
						 LPVOID	lpData)
{
	try
	{
		//Is it legal ?
		if (!pData ||
			!usSize)
		{
			//Report it
			ReportError("AddData","Recieved null data!");

			//Exit
			return;
		}

		//Perpare the data
		WaitingRequests aData;
		
		//Copy the memory
		aData.pData=new char[usSize];
		memcpy(aData.pData,
			   pData,
			   usSize);

		//Rest of the data
		aData.lpData=lpData;
		aData.usSize=usSize;

		//Lock first
		CCriticalAutoRelease aRelease(m_pCSectionDeque);
		
		//Insert it
		m_aPendingData.push_back(aData);
	}
	ERROR_HANDLER("AddData")
}

CDNSSocket::WaitingRequests CDNSSocket::GetData()const
{
	//Empty data
	WaitingRequests aEmptyData;
	aEmptyData.pData=NULL;
	aEmptyData.usSize=0;
	aEmptyData.lpData=NULL;
	
	try
	{
		//Lock first
		CCriticalAutoRelease aRelease(m_pCSectionDeque);

		//Do we have data
		if (m_aPendingData.empty())
			return aEmptyData;

		//Get the front
		WaitingRequests aData;
		aData=m_aPendingData.front();

		//Remove the front
		m_aPendingData.pop_front();

		//Exit
		return aData;
	}
	ERROR_HANDLER_RETURN("GetData",aEmptyData)
}

BOOL CDNSSocket::Flush()
{
	try
	{
		//Are we connected ?
		if (m_aConnectionStatus!=csConnected)
			return TRUE;

		//Get the data
		WaitingRequests aData;
		aData=GetData();

		//Do while we have data
		while (aData.pData)
		{
			//Protect the data
			CArray_ptr<char> pProtection(aData.pData);

			//Send it
			if (!m_pTCP->Send(aData.pData,
							  aData.usSize))
			{
				//Write the error
				ReportError("Flush","Failed to send!");

				//Get the error
				return FALSE;
			}

			//Can we continue ?
			if (m_aConnectionStatus!=csConnected)
				aData.pData=NULL;
			else
				aData=GetData();
		}
		
		//We are done
		return TRUE;
	}
	ERROR_HANDLER_RETURN("Flush",FALSE)
}

CDNSAnswers* CDNSSocket::Receive()
{
	if (m_bAsync)
		return NULL; //We will receive it later
	
	try
	{
		//Here we initialize the receive of the DNS
		char* pBuffer;
		pBuffer=new char[2000];

		//How much data did we receive
		int iReceived=0;

		if (m_bUseTCP)
			iReceived=m_pTCP->Receive(pBuffer,
									  2000);
		else
			iReceived=m_pUDP->Receive(pBuffer,
									  2000);

		//Check if we had errors
		if (iReceived==CSpoofBase::GetErrorCode() ||
			!iReceived)
		{
			//Delete the memory
			delete [] pBuffer;

			//Done
			return FALSE;
		}

		//Parse it
		return SocketReceive(pBuffer,
							 iReceived,
							 m_bUseTCP);
	}
	ERROR_HANDLER_RETURN("Receive",NULL)
}

BOOL CDNSSocket::SetConnectionTimeout(int iMS)
{
	try
	{
		//Save it
		m_iTimeout=iMS;

		//Which socket we use
		if (m_bUseTCP)
			return m_pTCP->SetConnectionTimeout(iMS);
		else
			return TRUE;
	}
	ERROR_HANDLER_RETURN("SetTimeout",FALSE)
}

CDNSAnswers* CDNSSocket::ParseAnswer(char* pBuffer,
									 char* pOriginalBuffer,
									 int iBufferLengh,
									 BOOL bFromThread,
									 BOOL bTCP)
{
	try
	{
		//Check if we have a thread manager
		if (m_pThreadManager && 
			!bFromThread)
		{
			//Check the buf len
			if (iBufferLengh<=0)
			{
				//Report it
				ReportError("ParseAnswer","Invalid buffer length!");

				//Exit
				return NULL;
			}

			//Create a job
			ParseThreadData* pData;
			pData=new ParseThreadData;

			//Set the data
			pData->pData=pBuffer;
			pData->pOriginalData=pOriginalBuffer;
			pData->iDataSize=iBufferLengh;
			pData->pSocket=this;
			pData->bTCP=bTCP;

			//Submit
			m_pThreadManager->SubmitJob(ThreadProc,
										(LPVOID)pData);

			//Exit
			return NULL;
		}

		//Parse the answer
		CDNSAnswers* pAnswers;
		pAnswers=new CDNSAnswers;

		//Try to parse
		int iResult;
		iResult=pAnswers->Parse(pBuffer,
								iBufferLengh);

		if (!iResult)
		{
			//Delete it, and set to null
			delete pAnswers;
			pAnswers=NULL;

			//Write the error
			ReportError("ParseAnswer","Failed to parse answers");

			//Delete the memory
			delete [] pOriginalBuffer;

			//And exit
			return NULL;
		}

		if (m_bAsync)
			//Inform the manager
			m_pFather->OnDNSReceive(pAnswers);

		if (iResult<iBufferLengh)
		{
			//Check if TCP
			if (bTCP)
				iResult+=2;

			//Change the data and do again
			iBufferLengh-=iResult;
			pBuffer+=iResult;

			//And parse again
			return ParseAnswer(pBuffer,
							   pOriginalBuffer,
							   iBufferLengh,
							   TRUE,
							   bTCP);
		}
		else
			delete [] pOriginalBuffer;

		//Done
		return pAnswers;
	}
	ERROR_HANDLER_RETURN("ParseAnswer",NULL)
}

void CDNSSocket::SocketError(int iErrorCode)
{
	try
	{
		//Relay the error
		m_pFather->OnDNSError(iErrorCode,
							  NULL);
	}
	ERROR_HANDLER("SocketError")
}

void CDNSSocket::SocketClosed()
{
	try
	{
		//Set to not connected
		SetConnectionStatus(csDisconnected);				
	}
	ERROR_HANDLER("SocketClosed")
}

CDNSAnswers* CDNSSocket::SocketReceive(char* pBuffer, 
									   int iBufferLengh,
									   BOOL bTCP)
{
	try
	{
		//Save the buffer
		char* pOriginalBuffer;
		pOriginalBuffer=pBuffer;

		//Check is it TCP
		if (bTCP)
		{
			//Adjust the buffer
			pBuffer+=2;
			iBufferLengh-=2;

			//Try to flush, if we have any ?
			if (!Flush())
				//Report it
				ReportError("SocketReceive","Failed to flush!");
		}

		//Parse the answer
		return ParseAnswer(pBuffer,
						   pOriginalBuffer,
						   iBufferLengh,
						   !m_bAsync,
						   bTCP);
	}
	ERROR_HANDLER_RETURN("SocketReceive",NULL)	
}

void CDNSSocket::SetMultithreaded(int iThreadNumber)
{
	try
	{
		//Do we have a thread manager
		if (m_pThreadManager)
			return;

		//Create it
		m_pThreadManager=new CThreadPool(iThreadNumber);

		//We own the thread manager
		m_bOwners=TRUE;
	}
	ERROR_HANDLER("SetMultithreaded")
}

void CDNSSocket::SetMultithreaded(CThreadPool* pThreadManager)
{
	try
	{
		//Delete an old manager, if we have it
		if (m_pThreadManager)
			delete m_pThreadManager;

		//Save it
		m_pThreadManager=pThreadManager;

		//We don't own the thread manager
		m_bOwners=FALSE;
	}
	ERROR_HANDLER("SetMultithreaded")
}

void CDNSSocket::ThreadProc(LPVOID lpParam)
{
	try
	{
		//Convert the param
		ParseThreadData* pData;
		pData=(ParseThreadData*)lpParam;

		//Parse it
		pData->pSocket->ParseAnswer(pData->pData,
									pData->pOriginalData,
									pData->iDataSize,
									TRUE,
									pData->bTCP);

		//Delete the data
		delete pData;
	}
	ERROR_HANDLER_STATIC(CDNSSocket_Class,"ThreadProc")
}

//of namespace
}