/*
 *  Copyright (c) 2000-2003 Barak Weichselbaum <barak@komodia.com>
 *  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * Contact info:
 * -------------
 *
 * Site:					http://www.komodia.com
 * Main contact:			barak@komodia.com
 * For custom projects, 
 * consulting, or other
 * paid services:			sales@komodia.com
 */

#include "stdafx.h"
#include "TCPPortScanner.h"

#include "ErrorHandlerMacros.h"
#include "OSManager.h"
#include "GenericCriticalSection.h"
#include "ThreadPool.h"
#include "GenericThread.h"

#ifdef _MEMORY_DEBUG 
	#define new	   DEBUG_NEW  
	#define malloc DEBUG_MALLOC  
    static char THIS_FILE[] = __FILE__;  
#endif

KOMODIA_NAMESPACE_START

// ----------------------------- CTCPScanner start ---------------------

#define CTCPScanner_Class "CTCPScanner"

CTCPPortScanner::CTCPScanner::CTCPScanner(CTCPPortScanner* pFather,
										  unsigned short usPort,
										  int iDataSize,
										  DWORD dwTimeout) : CTCPSocketAsync(FALSE),
															 m_pFather(pFather),
															 m_iDataSize(iDataSize),
															 m_dwTimeout(dwTimeout),
															 m_usPort(usPort)
{
	try
	{
		//Set our name
		SetName(CTCPScanner_Class);
	}
	ERROR_HANDLER("CTCPScanner")
}

CTCPPortScanner::CTCPScanner::~CTCPScanner()
{
}

BOOL CTCPPortScanner::CTCPScanner::OnSocketTimeout()
{
	try
	{
		//Kill the timer
		KillTimer();

		//Socket is done
		m_pFather->SocketDone(m_usPort,TRUE);

		//Delete ourselves
		delete this;

		//Done
		return FALSE;
	}
	ERROR_HANDLER_RETURN("OnSocketTimeout",FALSE)
}

BOOL CTCPPortScanner::CTCPScanner::OnSocketConnect(int iErrorCode)
{
	try
	{
		//Do we have an error
		if (iErrorCode || 
			!m_iDataSize)
		{
			//Notify father
			m_pFather->SocketDone(m_usPort,
								  iErrorCode==0);
			
			//Delete ourselves
			DeleteSocketFromThread(7000);

			//Done
			return TRUE;
		}

		//Set the timeout
		if (!SetTimeout(m_dwTimeout))
		{
			//Report it
			ReportError("OnSocketConnect","Failed to set timeout!");

			//Notify father
			m_pFather->SocketDone(m_usPort,TRUE);

			//Delete ourselves
			DeleteSocketFromThread(7000);
		}
		
		//Exit
		return TRUE;
	}
	ERROR_HANDLER_RETURN("OnSocketConnect",TRUE)
}

BOOL CTCPPortScanner::CTCPScanner::OnSocketClose(int iErrorCode)
{
	try
	{
		//Notify father
		m_pFather->SocketDone(m_usPort,TRUE);

		//Delete ourselves
		DeleteSocketFromThread();

		//Exit
		return TRUE;
	}
	ERROR_HANDLER_RETURN("OnSocketClose",TRUE)
}

BOOL CTCPPortScanner::CTCPScanner::OnSocketReceive(int iErrorCode)
{
	try
	{
		//Can we receive
		if (!m_iDataSize)
			return TRUE;

		//Kill the timer
		KillTimer();

		//Allocate the buffer
		char* pData;
		pData=new char[m_iDataSize];

		//Receive the data
		int iSize;
		iSize=Receive(pData,m_iDataSize-1);

		//Do we have an error ?
		if (iSize==GetErrorCode())
		{
			//Notify father
			m_pFather->SocketDone(m_usPort,
								  TRUE);

			//Delete the buffer
			delete [] pData;
		}
		else
		{
			//Set terminating null
			pData[iSize]=0;

			//Notify father
			m_pFather->SocketDone(m_usPort,
								  pData,
								  iSize);
		}

		//Delete ourselves
		DeleteSocketFromThread(7000);

		//Exit
		return TRUE;
	}
	ERROR_HANDLER_RETURN("OnSocketReceive",TRUE)
}

// ----------------------------- CTCPScanner end ---------------------

#define CTCPPortScanner_Class "CTCPPortScanner"

CTCPPortScanner::CTCPPortScanner() : CErrorHandler(),
									 m_pCSection(NULL),
									 m_bDone(TRUE),
									 m_iReceiveData(0),
									 m_iConnectionTimeout(0),
									 m_iMaxSockets(50),
									 m_bFinished(FALSE),
									 m_pCSectionDone(NULL)
{
	try
	{
		//Set our name
		SetName(CTCPPortScanner_Class);

		//Create the CS
		m_pCSection=COSManager::CreateCriticalSection();
		m_pCSectionDone=COSManager::CreateCriticalSection();

		//Zero the pairs
		m_aCurrentPair.usFromPort=0;
		m_aCurrentPair.usToPort=0;
	}
	ERROR_HANDLER("CTCPPortScanner")
}

CTCPPortScanner::~CTCPPortScanner()
{
	try
	{
		//Delete the CS
		delete m_pCSection;
		delete m_pCSectionDone;

		//Delete the port data
		DeletePortsMap();
	}
	ERROR_HANDLER("~CTCPPortScanner")
}

void CTCPPortScanner::AddPorts(unsigned short usFromPort,
							   unsigned short usToPort)
{
	try
	{
		//Check we can insert them
		if (!(usFromPort &&
			  usToPort &&
			  usFromPort<=usToPort))
			return;

		//Insert them
		ScanPair aData;
		aData.usFromPort=usFromPort;
		aData.usToPort=usToPort;

		//Lock
		CCriticalAutoRelease aRelease(m_pCSection,TRUE);

		m_aPortsToScan.push_back(aData);
	}
	ERROR_HANDLER("AddPorts")
}

BOOL CTCPPortScanner::IsDone()const
{
	return m_bDone;
}

void CTCPPortScanner::SetReceiveData(int iMaxDataSize,
									 DWORD dwTimeout)
{
	if (!m_bDone ||
		!m_bFinished)
		return;

	try
	{
		m_iReceiveData=iMaxDataSize;
		m_dwTimeout=dwTimeout;
	}
	ERROR_HANDLER("SetReceiveData")
}

DWORD CTCPPortScanner::GetReceiveDataTimeout()const
{
	return m_dwTimeout;
}

int CTCPPortScanner::GetReceiveDataSize()const
{
	return m_iReceiveData;
}

unsigned short CTCPPortScanner::GetNextPort()
{
	try
	{
		//Lock the data
		CCriticalAutoRelease aRelease(m_pCSection,TRUE);

		//Do we have a pair
		if (!m_aCurrentPair.usFromPort ||
			!m_aCurrentPair.usToPort ||
			m_aCurrentPair.usFromPort>m_aCurrentPair.usToPort)
			//Get a pair
			if (m_aPortsToScan.begin()==m_aPortsToScan.end())
				return 0;
			else
			{
				//Get the pair and remove it
				m_aCurrentPair=m_aPortsToScan.front();
				m_aPortsToScan.pop_front();
			}

		//Get the pair we want to send
		unsigned short usPort;
		usPort=m_aCurrentPair.usFromPort;

		//Increase the port
		++m_aCurrentPair.usFromPort;

		//Done
		return usPort;
	}
	ERROR_HANDLER_RETURN("GetNextPort",0)
}

BOOL CTCPPortScanner::CanScan()
{
	try
	{
		//Lock the data
		CCriticalAutoRelease aRelease(m_pCSection,TRUE);

		//Do we have a pair
		if (!m_aCurrentPair.usFromPort ||
			!m_aCurrentPair.usToPort ||
			m_aCurrentPair.usFromPort>m_aCurrentPair.usToPort)
			//Get a pair
			if (m_aPortsToScan.begin()==m_aPortsToScan.end())
				return FALSE;
			else
			{
				//Get the pair and remove it
				m_aCurrentPair=m_aPortsToScan.front();
				m_aPortsToScan.pop_front();
			}

		return TRUE;
	}
	ERROR_HANDLER_RETURN("CanScan",FALSE)
}

BOOL CTCPPortScanner::IsError()const
{
	return m_bError;
}

void CTCPPortScanner::SocketDone(unsigned short usPort,
								 BOOL bConnected)
{
	try
	{
		if (m_bDone ||
			m_bFinished)
			return;

		//Only if connected
		if (bConnected)
		{
			//Lock the data
			CCriticalAutoRelease aRelease(m_pCSection,TRUE);

			//Did we receive this answer already?
			if (m_aPortAnswers.find(usPort)!=m_aPortAnswers.end())
				return;

			//Add it to the list
			m_aPortList.insert(usPort);

			//And to the existing ports
			m_aPortAnswers.insert(usPort);
		}

		//Inicate for user
		OnPort(usPort,
			   bConnected);

		long lCount;
		lCount=AdjustNumberOfSockets(-1);

		//Decrease our count
		if (!CanScan() &&
			lCount<=0)
			ScanDone(FALSE);
		else
		{
			//Quit flag
			BOOL bQuit;
			bQuit=FALSE;

			//Create the socket
			while (!bQuit)
			{
				//Adjust the number of sockets
				AdjustNumberOfSockets(1);

				//Allocate it
				if (!(bQuit=NewSocket()))
					ReportError("SocketDone","Error creating the socket!");
			}
		}
	}
	ERROR_HANDLER("SocketDone")
}

void CTCPPortScanner::SocketDone(unsigned short usPort,
								 char* pData,
								 int iDataSize)
{
	try
	{
		if (m_bDone ||
			m_bFinished)
			return;

		{
			//Lock the data
			CCriticalAutoRelease aRelease(m_pCSection,TRUE);

			//Did we receive this answer already?
			if (m_aPortAnswers.find(usPort)!=m_aPortAnswers.end())
				return;

			//Add it to the list
			m_aPortList.insert(usPort);

			//And to the existing ports
			m_aPortAnswers.insert(usPort);

			//Add the data to the map
			m_aPortsData.insert(DataMap::value_type(usPort,pData));
		}

		//Inidicate for user
		OnPortData(usPort,
				   pData,
				   iDataSize);

		long lCount;
		lCount=AdjustNumberOfSockets(-1);

		//Decrease our count
		if (!CanScan() &&
			lCount<=0)
			ScanDone(FALSE);
		else
		{
			//Quit flag
			BOOL bQuit;
			bQuit=FALSE;

			//Create the socket
			while (!bQuit)
			{
				//Adjust the number of sockets
				AdjustNumberOfSockets(1);

				//Allocate it
				if (!(bQuit=NewSocket()))
					ReportError("SocketDone","Error creating the socket!");
			}
		}
	}
	ERROR_HANDLER("SocketDone")
}

void CTCPPortScanner::DeletePortsMap()
{
	try
	{
		//Get the iterator
		DataMap::iterator aIterator;
		aIterator=m_aPortsData.begin();

		//Start
		while (aIterator!=m_aPortsData.end())
		{
			//Save the iterator
			DataMap::iterator aBackupIterator;
			aBackupIterator=aIterator;

			//Advance it
			++aIterator;

			//Delete the data
			delete [] aBackupIterator->second;

			//Remove the data
			m_aPortsData.erase(aBackupIterator);
		}
	}
	ERROR_HANDLER("DeletePortsMap")
}

BOOL CTCPPortScanner::NewSocket()
{
	try
	{
		//Are we done
		if (m_bDone ||
			m_bFinished)
			return TRUE;

		//Get a port
		unsigned short usPort;
		usPort=GetNextPort();

		//Do we have it
		if (!usPort)
		{
			//Are we done
			if (AdjustNumberOfSockets(-1)<=0)
				ScanDone(FALSE);

			//Exit
			return TRUE;
		}

		//Create the socket
		CTCPSocketAsync* pSocket;
		pSocket=AllocateSocket(usPort);
		
		//Create it
		if (!pSocket)
		{
			//Are we done
			if (AdjustNumberOfSockets(-1)<=0 &&
				!CanScan())
			{
				//Run the scan
				ScanDone(FALSE);

				//Exit
				return TRUE;
			}
			else
				return FALSE;
		}

		//Connect it
		if (!pSocket->Connect(0,
							  m_aTarget,
							  usPort))
		{
			//Dispose the socket
			DestroySocket(pSocket);

			//Are we done
			if (AdjustNumberOfSockets(-1)<=0 &&
				!CanScan())
			{
				//Run the scan
				ScanDone(FALSE);

				//Exit
				return TRUE;
			}
			else
				return FALSE;
		}

		//We scanned the port
		PortScanned(usPort);

		//Done
		return TRUE;
	}
	ERROR_HANDLER_RETURN("NewSocket",FALSE)
}

int CTCPPortScanner::AdjustNumberOfSockets(int iIncrement)
{
	try
	{
		//Lock the data
		CCriticalAutoRelease aRelease(m_pCSection,TRUE);

		//Readjust
		m_iActiveSockets+=iIncrement;

		//Return it
		return m_iActiveSockets;
	}
	ERROR_HANDLER_RETURN("AdjustNumberOfSockets",0)
}

DWORD CTCPPortScanner::DoneThread(LPVOID lpData)
{
	try
	{
		//Get the structure
		std::auto_ptr<ThreadData> pData((ThreadData*)lpData);

		//Get the class
		CTCPPortScanner* pClass;
		pClass=(CTCPPortScanner*)pData->pScanner;

		//Notify library
		pClass->TCPScanDoneLib(pData->bError);

		//Set flags
		pClass->m_bDone=TRUE;
		pClass->m_bError=pData->bError;

		//Notify user
		pClass->TCPScanDone(pData->bError);

		//Done
		return TRUE;
	}
	ERROR_HANDLER_STATIC_RETURN(CTCPPortScanner_Class,"DoneThread",FALSE)
}

void CTCPPortScanner::ScanDone(BOOL bError,
							   BOOL bSpawnThread)
{
	try
	{
		{
			//Lock
			CCriticalAutoRelease aRelease(m_pCSectionDone);

			//What shell we do?
			if (m_bDone ||
				m_bFinished)
				return;

			//Set we are finished
			m_bFinished=TRUE;
		}

		//Thread's data
		ThreadData* pData;
		pData=new ThreadData;

		//Populate it
		pData->bError=bError;
		pData->pScanner=this;
		
		//Do we have a thread
		if (bSpawnThread)
		{
			//Spawn a thread
			CGenericThread* pThread;
			pThread=COSManager::CreateThread(DoneThread);

			//Set to autodelete
			pThread->SetAutoDelete(TRUE);

			//Start it
			pThread->Start((LPVOID)pData);
		}
		else
			DoneThread((LPVOID)pData);
	}
	ERROR_HANDLER("ScanDone")
}

const CTCPPortScanner::PortsList& CTCPPortScanner::GetPortList()const
{
	return m_aPortList;
}

void CTCPPortScanner::OnPort(unsigned short usPort,
							 BOOL bConnected)
{
}

void CTCPPortScanner::OnPortData(unsigned short usPort,
								 const char* pData,
								 int iDataSize)
{
}

void CTCPPortScanner::TCPScanDone(BOOL bError)
{
}

const char* CTCPPortScanner::GetData(unsigned short usPort)const
{
	try
	{
		//Try to look for the data
		DataMap::const_iterator aIterator;
		aIterator=m_aPortsData.find(usPort);

		//Do we have it
		if (aIterator!=m_aPortsData.end())
			return aIterator->second;
		else
			return NULL;
	}
	ERROR_HANDLER_RETURN("GetData",NULL)
}

char* CTCPPortScanner::GetDataForModify(unsigned short usPort)const
{
	try
	{
		//Try to look for the data
		DataMap::const_iterator aIterator;
		aIterator=m_aPortsData.find(usPort);

		//Do we have it
		if (aIterator!=m_aPortsData.end())
			return aIterator->second;
		else
			return NULL;
	}
	ERROR_HANDLER_RETURN("GetDataForModify",NULL)
}

void CTCPPortScanner::SetConnectionTimeout(int iMS)
{
	m_iConnectionTimeout=iMS;
}

int CTCPPortScanner::GetConnectionTimeout()const
{
	return m_iConnectionTimeout;
}

void CTCPPortScanner::SetMaxSockets(int iMaxSockets)
{
	m_iMaxSockets=iMaxSockets;
}

int CTCPPortScanner::GetMaxSockets()const
{
	return m_iMaxSockets;
}

void CTCPPortScanner::SetError(BOOL bError)
{
	m_bError=bError;
}

void CTCPPortScanner::SetDone(BOOL bDone)
{
	m_bDone=bDone;

	//Set the finish flag
	if (!bDone)
		m_bFinished=FALSE;
}

void CTCPPortScanner::ResetSocketCount()
{
	m_iActiveSockets=0;
}

void CTCPPortScanner::SetTarget(IP aTarget)
{
	m_aTarget=aTarget;
}

IP CTCPPortScanner::GetTarget()const
{
	return m_aTarget;
}

void CTCPPortScanner::ResetPair()
{
	m_aCurrentPair.usFromPort=0;
	m_aCurrentPair.usToPort=0;
}

void CTCPPortScanner::ResetData()
{
	try
	{
		//Clear the sets
		m_aPortAnswers.clear();
		m_aPortList.clear();
	}
	ERROR_HANDLER("ResetData")
}

CTCPSocketAsync* CTCPPortScanner::AllocateSocket(unsigned short usPort)const
{
	try
	{
		CTCPScanner* pSocket;
		pSocket=new CTCPScanner((CTCPPortScanner*)this,
								usPort,
								GetReceiveDataSize(),
								GetReceiveDataTimeout());

		//Do we have receive data ?
		if (!GetReceiveDataSize())
			//Disable receive
			pSocket->DisableEvents(CAsyncSocket::aeReceive);

		if (!pSocket->Create())
		{
			//Delete it
			delete pSocket;

			//Error
			return NULL;
		}
		else
		{
			//Set timeout
			if (GetConnectionTimeout())
				if (!pSocket->SetConnectionTimeout(GetConnectionTimeout()))
				{	
					//Delete it
					delete pSocket;

					//Error
					return NULL;
				}

			//Done
			return pSocket;
		}			
	}
	ERROR_HANDLER_RETURN("AllocateSocket",NULL)
}

void CTCPPortScanner::DestroySocket(CTCPSocketAsync* pSocket)const
{
	try
	{
		delete pSocket;
	}
	ERROR_HANDLER("DestroySocket")
}

void CTCPPortScanner::TCPScanDoneLib(BOOL bError)
{
}

void CTCPPortScanner::PortScanned(unsigned short usPort)
{
}

BOOL CTCPPortScanner::IsFinished()const
{
	return m_bFinished;
}

BOOL CTCPPortScanner::StopScan()
{
	try
	{
		//Call we are done
		ScanDone(TRUE,
				 FALSE);

		//Done
		return TRUE;
	}
	ERROR_HANDLER_RETURN("StopScan",FALSE)
}

KOMODIA_NAMESPACE_END
