/*
 *  Copyright (c) 2000-2003 Barak Weichselbaum <barak@komodia.com>
 *  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * Contact info:
 * -------------
 *
 * Site:					http://www.komodia.com
 * Main contact:			barak@komodia.com
 * For custom projects, 
 * consulting, or other
 * paid services:			sales@komodia.com
 */

#include "stdafx.h"
#include "TCPPortScannerMultithreaded.h"

#include "ErrorHandlerMacros.h"
#include "ThreadPool.h"

#ifdef _MEMORY_DEBUG 
	#define new	   DEBUG_NEW  
	#define malloc DEBUG_MALLOC  
    static char THIS_FILE[] = __FILE__;  
#endif

KOMODIA_NAMESPACE_START

#define CTCPPortScannerMultithreaded_Class "CTCPPortScannerMultithreaded"

CTCPPortScannerMultithreaded::CTCPPortScannerMultithreaded() : CTCPPortScanner(),
															   m_pThreadPool(NULL),
															   m_iMaxThreads(10)
{
	try
	{
		//Set our name
		SetName(CTCPPortScannerMultithreaded_Class);
	}
	ERROR_HANDLER("CTCPPortScannerMultithreaded")
}

CTCPPortScannerMultithreaded::~CTCPPortScannerMultithreaded()
{
	try
	{
		//Delete the pool
		if (m_pThreadPool &&
			m_bOwners)
			delete m_pThreadPool;
	}
	ERROR_HANDLER("~CTCPPortScannerMultithreaded")
}

BOOL CTCPPortScannerMultithreaded::Scan(const std::string& rDestinationAddress)
{
	try
	{
		return Scan(CSpoofBase::StringToLong(rDestinationAddress));
	}
	ERROR_HANDLER_RETURN("Scan",FALSE)
}

BOOL CTCPPortScannerMultithreaded::Scan(IP aTarget)
{
	//Check max sockets are ok
	if (GetMaxSockets()<=0 ||
		!IsDone() ||
		!aTarget ||
		(m_iMaxThreads<=0 &&
		 !m_pThreadPool))
	{
		//Report it
		ReportError("Scan","Invalid parameters!");

		//Exit
		return FALSE;
	}

	try
	{
		//Reset the pair
		ResetPair();

		//Can we scan
		if (!CanScan())
		{
			//Error
			ReportError("Scan","No ports given!");

			//Error
			SetError(TRUE);

			//Exit
			return FALSE;
		}

		//Set the flags
		SetError(FALSE);
		SetDone(FALSE);

		//Save the number of max sockets
		ResetSocketCount();

		//Delete the port data
		DeletePortsMap();

		//Delete other data
		ResetData();

		//Set our target
		SetTarget(aTarget);

		//Do we need to modify the thread pool
		if (m_bOwners &&
			m_pThreadPool &&
			m_iMaxThreads>0 &&
			m_pThreadPool->GetMaxThreads()!=m_iMaxThreads)
		{
			//Delete it for recreation
			delete m_pThreadPool;
			m_pThreadPool=NULL;
		}
	
		//Thread pool, do we have it
		if (!m_pThreadPool)
		{
			//Create it
			m_pThreadPool=new CThreadPool(m_iMaxThreads);

			//We are the owners
			m_bOwners=TRUE;
		}

		//Start the loop
		for (int iCounter=0;
			 iCounter<GetMaxSockets();
			 ++iCounter)
			//Spawn a thread
			m_pThreadPool->SubmitJob(SocketProc,
									 (LPVOID)this);

		//Done
		return TRUE;
	}
	ERROR_HANDLER_RETURN("Scan",FALSE)
}

void CTCPPortScannerMultithreaded::SocketProc(LPVOID lpData)
{
	try
	{
		//Get the class
		CTCPPortScannerMultithreaded* pClass;
		pClass=(CTCPPortScannerMultithreaded*)lpData;

		//Quit flag
		BOOL bQuit;
		bQuit=FALSE;

		//Create the socket
		while (!bQuit)
		{
			//Adjust the number of sockets
			pClass->AdjustNumberOfSockets(1);

			//Allocate it
			if (!(bQuit=pClass->NewSocket()))
				ReportStaticError(CTCPPortScannerMultithreaded_Class,"SocketProc","Error creating the socket!");
		}
	}
	ERROR_HANDLER_STATIC(CTCPPortScannerMultithreaded_Class,"SocketProc")
}

void CTCPPortScannerMultithreaded::SetMultithreadedPool(CThreadPool* pPool)
{
	try
	{
		//Do we have the old pool
		if (m_pThreadPool &&
			m_bOwners)
			delete m_pThreadPool;

		//Save this one
		m_pThreadPool=pPool;

		//We are not the owners
		m_bOwners=FALSE;
	}
	ERROR_HANDLER("SetMultithreadedPool")
}

void CTCPPortScannerMultithreaded::SetMaxThreads(int iThreads)
{
	m_iMaxThreads=iThreads;
}

int CTCPPortScannerMultithreaded::GetMaxThreads()const
{
	return m_iMaxThreads;
}

KOMODIA_NAMESPACE_END
