/*
 *  Copyright (c) 2000-2003 Barak Weichselbaum <barak@komodia.com>
 *  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * Contact info:
 * -------------
 *
 * Site:					http://www.komodia.com
 * Main contact:			barak@komodia.com
 * For custom projects, 
 * consulting, or other
 * paid services:			sales@komodia.com
 */

#include "stdafx.h"
#include "SpoofBase.h"

#include "ErrorHandlerMacros.h"
#include "OSManager.h"
#include "GenericCriticalSection.h"
#include "GenericEvent.h"
#include "ManagedThread.h"
#include "LibConfig.h"

#include <memory>

#ifdef _MEMORY_DEBUG 
	#define new	   DEBUG_NEW  
	#define malloc DEBUG_MALLOC  
    static char THIS_FILE[] = __FILE__;  
#endif

KOMODIA_NAMESPACE_START

//Are we multithreaded?
BOOL CSpoofBase::m_bMultiThreaded=FALSE;

//Are we initialized?
BOOL CSpoofBase::m_bInitialized=FALSE;

//Our number of threads
int CSpoofBase::m_ulNumberOfThreads=0;

//Our global CS
CGenericCriticalSection* CSpoofBase::m_pCSection=NULL;

//Our global shutdown class
CSpoofBase* CSpoofBase::m_pShutdownClass=NULL;

//Our managed thread
CManagedThread* CSpoofBase::m_pThread=NULL;

//Our wait timeout
#define THREAD_TIMEOUT 10000

#define CSpoofBase_Class "CSpoofBase"

CSpoofBase::CSpoofBase() : CErrorHandler()
{
	try
	{
		//Set it
		SetName(CSpoofBase_Class);
	}
	ERROR_HANDLER("CSpoofBase")
}

CSpoofBase::CSpoofBase(const CSpoofBase& rBase) : CErrorHandler(rBase)
{
	try
	{
		//Set it
		SetName(CSpoofBase_Class);
	}
	ERROR_HANDLER("CSpoofBase")
}

CSpoofBase::~CSpoofBase()
{
}

void CSpoofBase::SetLastError(const std::string& rMethod) const
{
	try
	{
#ifdef WIN32
		//First set the error
		m_iLastError=WSAGetLastError();
#else
		m_iLastError=errno();
#endif

		//Check if there is an error
		if (m_iLastError)
			ReportError(rMethod,
						m_iLastError);
	}
	ERROR_HANDLER("SetLastError")
}

void CSpoofBase::SetLastError(const std::string& rMethod,
							  int iErrorCode)const
{
	try
	{
		//First set the error
		m_iLastError=iErrorCode;

		//Check if there is an error
		if (m_iLastError)
			ReportError(rMethod,
						m_iLastError);
	}
	ERROR_HANDLER("SetLastError")
}

int CSpoofBase::GetLastError()const
{
	return m_iLastError;
}

DWORD CSpoofBase::InitProc(CGenericEvent* pExitEvent,
						   LPVOID lpData)
{
	try
	{
		//Our data
		std::auto_ptr<ThreadData> pThreadData((ThreadData*)lpData);

		//Try to initialize the sockets
		BOOL bResult;
		bResult=InitializeSockets(pThreadData->bMultiThreaded,
								  pThreadData->ulNumberOfThreads);

		//Signal the event
		pThreadData->pEvent->Set();

		//Do we need to run a loop?
		if (bResult)
		{
			//Yes run a message loop
			MSG msg;

			//Set the message map to zeros
			memset(&msg,
				   0,
				   sizeof(msg));

			//Stop flag
			BOOL bRun;
			bRun=TRUE;

			//Start the loop
			while (bRun &&
				   pExitEvent->Wait(0))
			{
				while ((bRun=pExitEvent->Wait(0)) &&
					   PeekMessage(&msg, 
								   NULL, 
								   0, 
								   0, 
								   PM_REMOVE))
				{
					TranslateMessage(&msg);
					DispatchMessage(&msg); 
				}

				//Do we need to sleep?
				if (bRun)
					//No messages
					bRun=pExitEvent->Wait(1);
			}
		}

		//Done
		return TRUE;
	}
	ERROR_HANDLER_STATIC_RETURN(CSpoofBase_Class,"InitProc",FALSE)
}

BOOL CSpoofBase::InitializeSocketsNoMap(BOOL bMultiThreaded,
									    unsigned long ulNumberOfThreads)
{
	//To avoid double initialize
	if (m_bInitialized)
	{
		//Report it
		ReportStaticError(CSpoofBase_Class,"InitializeSocketsNoMap","Already initialized!");

		//Exit
		return TRUE;
	}

	//Check that the number of threads are OK?
	if (ulNumberOfThreads>CLibConfig::GetInstance().GetMaxThreads())
	{
		//Report it
		ReportStaticError(CSpoofBase_Class,"InitializeSocketsNoMap","Too many threads!");

		//Exit
		return FALSE;
	}

	//Do we have threads at all
	if (bMultiThreaded &&
		!ulNumberOfThreads)
	{
		//Report it
		ReportStaticError(CSpoofBase_Class,"InitializeSocketsNoMap","Didn't receive any threads!");

		//Exit
		return FALSE;
	}

	try
	{
		//Create the thread data
		ThreadData* pThreadData;
		pThreadData=new ThreadData;

		//Populate the data
		pThreadData->pEvent=COSManager::CreateEvent();
		pThreadData->bMultiThreaded=bMultiThreaded;
		pThreadData->ulNumberOfThreads=ulNumberOfThreads;

		//Create the thread
		m_pThread=new CManagedThread(InitProc);
		m_pThread->Start((LPVOID)pThreadData);

		//Wait on the event
		if (pThreadData->pEvent->Wait(THREAD_TIMEOUT))
		{
			//Report it
			ReportStaticError(CSpoofBase_Class,"InitializeSocketsNoMap","Timeout waiting for thread");

			//Delete the thread
			delete m_pThread;
			m_pThread=NULL;

			//Exit
			return FALSE;
		}

		//Are we initialized
		if (!IsInitialized())
		{
			//Delete the thread
			delete m_pThread;
			m_pThread=NULL;
		}

		//Done
		return IsInitialized();
	}
	ERROR_HANDLER_STATIC_RETURN(CSpoofBase_Class,"InitializeSocketsNoMap",FALSE)
}

BOOL CSpoofBase::InitializeSockets(BOOL bMultiThreaded,
								   unsigned long ulNumberOfThreads)
{
	//To avoid double initialize
	if (m_bInitialized)
	{
		//Report it
		ReportStaticError(CSpoofBase_Class,"InitializeSockets","Already initialized!");

		//Exit
		return TRUE;
	}

	//Check that the number of threads are OK?
	if (ulNumberOfThreads>CLibConfig::GetInstance().GetMaxThreads())
	{
		//Report it
		ReportStaticError(CSpoofBase_Class,"InitializeSockets","Too many threads!");

		//Exit
		return FALSE;
	}

	//Do we have threads at all
	if (bMultiThreaded &&
		!ulNumberOfThreads)
	{
		//Report it
		ReportStaticError(CSpoofBase_Class,"InitializeSockets","Didn't receive any threads!");

		//Exit
		return FALSE;
	}

	try
	{
		//Initialize the sockets
		WORD wVersionRequested;
		wVersionRequested=MAKEWORD(2,2);
 
		//Try to initialize
		WSADATA wsaData;
		int iErr;
		iErr=WSAStartup(wVersionRequested, 
						&wsaData);

		//Did we succeed?
		if (iErr!=0)
			/* Tell the user that we could not find a usable */
			/* WinSock DLL.                                  */
			return FALSE;
 
		/* Confirm that the WinSock DLL supports 2.2.*/
		/* Note that if the DLL supports versions greater    */
		/* than 2.2 in addition to 2.2, it will still return */
		/* 2.2 in wVersion since that is the version we      */
		/* requested.                                        */
 
		if (LOBYTE(wsaData.wVersion)!=2 || 
			HIBYTE(wsaData.wVersion)!=2)
		{
			/* Tell the user that we could not find a usable */
			/* WinSock DLL.                                  */
			WSACleanup();

			//Exit
			return FALSE;
		}

		//Save the threading information
		m_bMultiThreaded=bMultiThreaded;
		m_ulNumberOfThreads=ulNumberOfThreads;

		//Create the critical section
		m_pCSection=COSManager::CreateCriticalSection();

		//And we are initialized
		m_bInitialized=TRUE;

		return TRUE;
	}
	ERROR_HANDLER_STATIC_RETURN(CSpoofBase_Class,"InitializeSockets",FALSE)
}

BOOL CSpoofBase::ShutdownSockets()
{
	//Only if initialized
	if (!m_bInitialized)
		return TRUE;

	try
	{
		//Do we have a thread?
		if (m_pThread)
		{
			//Delete it
			delete m_pThread;
			m_pThread=NULL;
		}

		//Delete the CS
		delete m_pCSection;
		m_pCSection=NULL;

		//Notify shutdown class
		if (m_pShutdownClass)
		{
			m_pShutdownClass->NotifyShutdown();
			delete m_pShutdownClass;
		}

		if (WSACleanup()==GetErrorCode())
			return FALSE;

		//Not initialized anymore
		m_bInitialized=FALSE;

		//Done
		return TRUE;
	}
	ERROR_HANDLER_STATIC_RETURN(CSpoofBase_Class,"ShutdownSockets",FALSE)
}

void CSpoofBase::NotifyShutdown()
{
}

void CSpoofBase::RegisterShutdown(CSpoofBase* pBase)
{
	try
	{
		//Check if we already have a class
		if (m_pShutdownClass)
			delete m_pShutdownClass;

		//Take it
		m_pShutdownClass=pBase;
	}
	ERROR_HANDLER("RegisterShutdown")
}

std::string CSpoofBase::LongToStdString(unsigned long ulAddr)
{
	try
	{
		//First create the address
		in_addr addr;

		//Assign it
		addr.S_un.S_addr=ulAddr;

		//Enter the critical section
		CCriticalAutoRelease aRelease(m_pCSection);

		//Return the value
		return inet_ntoa(addr);
	}
	ERROR_HANDLER_STATIC_RETURN(CSpoofBase_Class,"LongToStdString","0.0.0.0")
}

char FAR * CSpoofBase::LongToString(unsigned long ulAddr)
{
	try
	{
		//First create the address
		in_addr addr;

		//Assign it
		addr.S_un.S_addr=ulAddr;

		//Enter the critical section
		CCriticalAutoRelease aRelease(m_pCSection);

		//Return the value
		return inet_ntoa(addr);
	}
	ERROR_HANDLER_STATIC_RETURN(CSpoofBase_Class,"LongToString",NULL)
}

BOOL CSpoofBase::IsMultiThreaded()
{
	return m_bMultiThreaded;
}

unsigned long CSpoofBase::GetNumberOfThreads()
{
	return m_ulNumberOfThreads;
}

BOOL CSpoofBase::IsInitialized()
{
	return m_bInitialized;
}

unsigned long CSpoofBase::StringToLong(const std::string& rAddress)
{
	return inet_addr(rAddress.c_str());
}

int CSpoofBase::GetSystemLastError() const
{
	return WSAGetLastError();
}

CGenericCriticalSection* CSpoofBase::GetGlobalCriticalSection()
{
	return m_pCSection;
}

int CSpoofBase::GetErrorCode()
{
	return SOCKET_ERROR;
}

KOMODIA_NAMESPACE_END