/*
 *  Copyright (c) 2000-2003 Barak Weichselbaum <barak@komodia.com>
 *  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * Contact info:
 * -------------
 *
 * Site:					http://www.komodia.com
 * Main contact:			barak@komodia.com
 * For custom projects, 
 * consulting, or other
 * paid services:			sales@komodia.com
 */

#include "stdafx.h"
#include "TCPSocketAsync.h"

#include "ErrorHandlerMacros.h"

#include "BlockedBuffer.h"

#ifdef _MEMORY_DEBUG 
	#define new	   DEBUG_NEW  
	#define malloc DEBUG_MALLOC  
    static char THIS_FILE[] = __FILE__;  
#endif

KOMODIA_NAMESPACE_START

#define CTCPSocketAsync_Class "CTCPSocketAsync"

CTCPSocketAsync::CTCPSocketAsync(BOOL bRawSocket) : CTCPSocket(bRawSocket), 
													CAsyncSocket(),
													m_bDisabledConnect(FALSE),
													m_ulTimeout(0)
{
	try
	{
		//Set our name
		SetName(CTCPSocketAsync_Class);
	}
	ERROR_HANDLER("CTCPSocketAsync")
}

CTCPSocketAsync::~CTCPSocketAsync()
{
	try
	{
		//We need to close it here
		Close();
	}
	ERROR_HANDLER("~CTCPSocketAsync")
}

BOOL CTCPSocketAsync::Connect(IP aDestinationAddress,
							  unsigned short usDestinationPort,
							  BOOL bDisableAsync,
							  BOOL bForceErrorEvent)
{
	try
	{
		//Delegate the call
		return Connect(0,
					   aDestinationAddress,
					   usDestinationPort,
					   bDisableAsync,
					   bForceErrorEvent);
	}
	ERROR_HANDLER_RETURN("Connect",FALSE)
}

BOOL CTCPSocketAsync::Connect(const std::string& rDestinationAddress,
							  unsigned short usDestinationPort,
							  BOOL bDisableAsync,
							  BOOL bForceErrorEvent)
{
	try
	{
		//Delegate the call
		return Connect(0,
					   rDestinationAddress,
					   usDestinationPort,
					   bDisableAsync,
					   bForceErrorEvent);
	}
	ERROR_HANDLER_RETURN("Connect",FALSE)
}

BOOL CTCPSocketAsync::Connect(unsigned short usSourcePort,
							  IP aDestinationAddress,
							  unsigned short usDestinationPort,
							  BOOL bDisableAsync,
							  BOOL bForceErrorEvent)
{
	try
	{
		//Quit if not ok
		if (!CheckSocketValid())
			return FALSE;

		//Set the async notification
		if (!bDisableAsync)
		{
			int iResult;
			iResult=InternalWSAAsyncSelect(WM_SOCKET_CONNECT,
										   FD_CONNECT);

			if (iResult)
			{
				//Get the error code
				int iErrorCode;
				iErrorCode=GetSystemLastError();

				//Report it
				SetLastError("Connect");

				//Do we need to call event?
				if (bForceErrorEvent)
					SocketConnected(iErrorCode);

				//Exit
				return FALSE;
			}

			//Set our timeout
			if (m_ulTimeout &&
				!IsBlocking())
				if (!SetSystemTimeout(m_ulTimeout))
				{
					//Report it
					ReportError("Connect","Failed to set timer!");

					//Do we need to call event?
					if (bForceErrorEvent)
						SocketConnected(GetErrorCode());

					//Exit
					return FALSE;
				}
		}
		//Set to non blocking!
		else if (!Block())
			return FALSE;

		//Set the flag
		m_bDisabledConnect=bDisableAsync;

		//Call the original connect
		BOOL bResult;
		bResult=CTCPSocket::Connect(usSourcePort,
									aDestinationAddress,
									usDestinationPort);

		//Reset the flag
		m_bDisabledConnect=FALSE;

		if (bResult)
		{
			//Call event, but only if in async
			if (!bDisableAsync &&
				!IsBlocking())
				//Call user, will add socket automatically
				return SocketConnected(0);
			else
				//Set as async
				return SetAsync();
		}
		else if (GetSystemLastError()!=WSAEWOULDBLOCK ||
				 bDisableAsync ||
				 IsBlocking())
		{
			if (m_ulTimeout)
				//Kill the timer
				KillSystemTimer();

			//Get the error code
			int iErrorCode;
			iErrorCode=GetSystemLastError();

			//Report it
			SetLastError("Connect");

			//Do we need to call event?
			if (bForceErrorEvent &&
				!bDisableAsync)
				SocketConnected(iErrorCode);

			//Exit
			return FALSE;
		}
		else
			return TRUE;
	}
	ERROR_HANDLER_RETURN("Connect",FALSE)
}

BOOL CTCPSocketAsync::Connect(unsigned short usSourcePort,
							  const std::string& rDestinationAddress,
							  unsigned short usDestinationPort,
							  BOOL bDisableAsync,
							  BOOL bForceErrorEvent)
{
	try
	{
		//Quit if not ok
		if (!CheckSocketValid())
			return FALSE;

		//Delegate the call
		return Connect(usSourcePort,
					   StringToLong(rDestinationAddress),
					   usDestinationPort,
					   bDisableAsync,
					   bForceErrorEvent);
	}
	ERROR_HANDLER_RETURN("Connect",FALSE)
}

BOOL CTCPSocketAsync::Listen(unsigned long ulBackLog)
{
	try
	{
		//Quit if not ok
		if (!CheckSocketValid())
			return FALSE;

		//Try to switch the mode
		int iResult;
		iResult=InternalWSAAsyncSelect(WM_SOCKET_ACCEPT,
									   FD_ACCEPT);

		//Is it OK
		if (iResult)
		{
			//Report it
			SetLastError("Listen");

			//Exit
			return FALSE;
		}

		//Delegate the call
		return CTCPSocket::Listen(ulBackLog);
	}
	ERROR_HANDLER_RETURN("Listen",FALSE)
}

BOOL CTCPSocketAsync::SetAsync()
{
	try
	{
		//Quit if not ok
		if (!CheckSocketValid())
			return FALSE;

		//Set event to read / write / close / oob
		int iResult;
		iResult=InternalWSAAsyncSelect(WM_SOCKET_GENERAL,
									   FD_WRITE | FD_READ | FD_CLOSE | FD_OOB);

		//What happend
		if (iResult)
		{
			//Report it
			SetLastError("SetAsync");

			//Exit
			return FALSE;
		}
		
		//Done
		return TRUE;
	}
	ERROR_HANDLER_RETURN("SetAsync",FALSE)
}

BOOL CTCPSocketAsync::SocketConnected(int iErrorCode)
{
	try
	{
		if (m_ulTimeout)
			//Kill the timer
			KillSystemTimer();

		//Did we got an error
		if (!iErrorCode)
			//Set the connection status
			SetConnectionStatus(TRUE);
		
		//Call father
		return CAsyncSocket::SocketConnected(iErrorCode);
	}
	ERROR_HANDLER_RETURN("SocketConnected",FALSE)
}

BOOL CTCPSocketAsync::SocketClosed(int iErrorCode)
{
	try
	{
		//We are not connected
		SetConnectionStatus(FALSE);

		//Call father
		return CAsyncSocket::SocketClosed(iErrorCode);
	}
	ERROR_HANDLER_RETURN("SocketClosed",FALSE)
}

void CTCPSocketAsync::Accepted()
{
	try
	{
		//Add to async list
		AddSocketToList();

		//Go to async mode
		SetAsync();
	}
	ERROR_HANDLER("CTCPSocketAsync")
}

SOCKET CTCPSocketAsync::GetAsyncHandle() const
{
	return GetHandle();
}

BOOL CTCPSocketAsync::Create()
{
	try
	{
		if (!CTCPSocket::Create())
			return FALSE;
		else
		{
			//New async settings
			SocketCreated();

			//Add to list
			AddSocketToList();

			//Done
			return TRUE;
		}
	}
	ERROR_HANDLER_RETURN("Create",FALSE)
}

BOOL CTCPSocketAsync::Close()
{
	try
	{
		//Quit if not ok
		if (!ValidSocket())
			return FALSE;

		//Kill the timer
		CAsyncSocket::SocketClosing();

		//Remove from socket list
		RemoveSocketFromList();

		//Done
		return CTCPSocket::Close();
	}
	ERROR_HANDLER_RETURN("Close",FALSE)
}

BOOL CTCPSocketAsync::IsAsyncClass() const
{
	try
	{
		//Check if we are blocking
		if (IsBlocking() || 
			m_bDisabledConnect)
			return FALSE;
		else
			//Not blocking
			return TRUE;
	}
	ERROR_HANDLER_RETURN("IsAsyncClass",FALSE)
}

void CTCPSocketAsync::BeforeAccept()
{
	try
	{
		//Remove ourselves
		RemoveSocketFromList();
	}
	ERROR_HANDLER("BeforeAccept")
}

BOOL CTCPSocketAsync::SendBlockedBuffer(const CBlockedBuffer::CBlockedData& rData)
{
	try
	{
		//Try to send it
		return SendNoAdd(rData.GetData(),
						 rData.GetDataSize())!=GetErrorCode();
	}
	ERROR_HANDLER_RETURN("SendBlockedBuffer",FALSE)
}

int CTCPSocketAsync::Send(const char *pBuffer,
						  unsigned long ulBufferLength)
{
	try
	{
		//Send the data
		int iResult;
		iResult=SendNoAdd(pBuffer,
						  ulBufferLength);

		//Did we succeed ?
		if (iResult==GetErrorCode())
			//Is it blocked send ?
			if (WSAGetLastError()==WSAEWOULDBLOCK &&
				GetBlockedBuffer())
			{
				//Add to the buffer, if we have one
				GetBlockedBuffer()->AddRecord(CBlockedBuffer::CBlockedData(pBuffer,
																		   ulBufferLength));

				//We have not error
				iResult=0;
			}
			//Blocking error (not really an error)
			else if (WSAGetLastError()==WSAEWOULDBLOCK)
				return 0;
			else
				//Set the error code
				SetLastError("Send");

		//Done
		return iResult;
	}
	ERROR_HANDLER_RETURN("Send",GetErrorCode())
}

int CTCPSocketAsync::SendNoAdd(const char* pBuffer,
							   unsigned long ulBufferLength)
{
	try
	{
		//Quit if not ok
		if (!CheckSocketValid())
			return FALSE;

		//Send the data
		int iResult;

		//And send it
		iResult=send(GetHandle(),
					 pBuffer,
					 ulBufferLength,
					 NULL);

		//And exit
		return iResult;
	}
	ERROR_HANDLER_RETURN("SendNoAdd",GetErrorCode())
}

BOOL CTCPSocketAsync::SocketWrite(int iErrorCode)
{
	try
	{
		//Only if we have a buffer
		if (GetBlockedBuffer())
			GetBlockedBuffer()->SendData();

		//Done
		return OnSocketWrite(iErrorCode);
	}
	ERROR_HANDLER_RETURN("OnSocketWrite",TRUE)
}

BOOL CTCPSocketAsync::SetConnectionTimeout(unsigned long ulMS)
{
	try
	{
		//Do we have a timeout ?
		if (HasSystemTimer())
			if (!KillSystemTimer())
			{
				//Report it
				ReportError("SetConnectionTimeout","Failed to kill previous timer!");

				//Exit
				return FALSE;
			}

		//Create the timer
		m_ulTimeout=ulMS;

		//Done
		return TRUE;
	}
	ERROR_HANDLER_RETURN("SetConnectionTimeout",FALSE)
}

void CTCPSocketAsync::SystemTimeout()
{
	//Call socket timeout
	SocketConnected(CONNECTION_TIMEOUT);
}

int CTCPSocketAsync::Receive(char* pBuffer,
							 unsigned long ulBufferLength)
{
	try
	{
		//Try to receive
		int iResult;
		if ((iResult=CTCPSocket::Receive(pBuffer,
										 ulBufferLength))>=0)
			return iResult;
		//Check is it an error
		else if (GetSystemLastError()==WSAEWOULDBLOCK)
			return 0;
		else
			return GetErrorCode();
	}
	ERROR_HANDLER_RETURN("Receive",GetErrorCode())
}

KOMODIA_NAMESPACE_END